from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import TransformationYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def transformation_destination_cases() -> Iterable:
    transformation_destination = [
        {
            "DataModelSource": {
                "externalId": "tr_first_transformation",
                "name": "example:first:transformation",
                "ignoreNullFields": True,
                "destination": {
                    "type": "instances",
                    "dataModel": {
                        "externalId": "my_data_model",
                        "version": "1",
                        "space": "my_space",
                        "destinationType": "my_view",
                    },
                    "instanceSpace": "my_instance_space",
                },
            }
        },
        {
            "DataModelSource": {
                "externalId": "tr_first_transformation",
                "name": "example:first:transformation",
                "ignoreNullFields": True,
                "destination": {
                    "type": "instances",
                    "dataModel": {
                        "externalId": "my_data_model",
                        "version": "1",
                        "space": "my_space",
                        "destinationType": "my_view",
                    },
                },
            }
        },
        {
            "ViewDataSource": {
                "externalId": "tr_first_transformation",
                "name": "example:first:transformation",
                "ignoreNullFields": True,
                "destination": {
                    "type": "nodes",
                    "view": {"externalId": "my_view", "version": "1", "space": "my_space"},
                    "edgeType": {"externalId": "my_edge_type", "space": "my_space"},
                },
            }
        },
        {
            "RawDataSource": {
                "externalId": "tr_first_transformation",
                "name": "example:first:transformation",
                "ignoreNullFields": True,
                "destination": {"type": "raw", "database": "my_database", "table": "my_table"},
            }
        },
        {
            "SequenceRowDataSource": {
                "externalId": "tr_first_transformation",
                "name": "example:first:transformation",
                "ignoreNullFields": True,
                "destination": {"type": "sequence_rows", "externalId": "my_sequence"},
            }
        },
    ]
    yield from (pytest.param(next(iter(des.values())), id=next(iter(des.keys()))) for des in transformation_destination)


def invalid_transformation_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "my_transformation", "ignoreNullFields": True},
        {"Missing required field: 'externalId'"},
        id="Missing externalId",
    )
    yield pytest.param(
        {
            "conflictMode": "upsert",
            "dataSetExternalId": "ops:001:monitor",
            "destination": {"type": "raw", "database": "ops:001:monitor:rawtable", "table": "rawtable"},
            "authentication.write": {
                "cdfProjectName": "my_project",
                "clientId": "my_client_id",
                "clientSecret": "my_client_secret",
                "scopes": ["USER_MANAGEMENT", "DATASETS_WRITE"],
                "tokenUri": "https://api.cognitedata.com/api/v1/oauth/token",
            },
            "externalId": "cdf_ops:raw:upsert:001:monitor",
            "ignoreNullFields": True,
            "isPublic": True,
            "name": "cdf_ops:raw:upsert:001:monitor",
            "authentication.read": {
                "cdfProjectName": "my_project",
                "clientId": "my_client_id",
                "clientSecret": "my_client_secret",
                "scopes": ["USER_MANAGEMENT", "DATASETS_READ"],
                "tokenUri": "https://api.cognitedata.com/api/v1/oauth/token",
            },
        },
        {"Unused field: 'authentication.read'", "Unused field: 'authentication.write'"},
        id="Invalid authentication - base on real use case",
    )
    yield pytest.param(
        {
            "externalId": "tr_invalid_type",
            "name": "Invalid ignore_null_fields",
            "ignoreNullFields": "yes",
        },
        {"In field ignoreNullFields input should be a valid boolean. Got 'yes' of type str."},
        id="Invalid ignore_null_fields type",
    )
    yield pytest.param(
        {
            "externalId": "tr_too_many_tags",
            "name": "Too many tags",
            "ignoreNullFields": True,
            "tags": ["a", "b", "c", "d", "e", "f"],
        },
        {"In field tags list should have at most 5 items after validation, not 6"},
        id="Too many tags in transformation",
    )
    yield pytest.param(
        {
            "externalId": "tr_invalid_conflict",
            "name": "Invalid conflict_mode",
            "ignoreNullFields": True,
            "conflictMode": "replace",
        },
        {"In field conflictMode input should be 'abort', 'delete', 'update' or 'upsert'. Got 'replace'."},
        id="Invalid conflict_mode value",
    )
    yield pytest.param(
        {
            "externalId": "tr_invalid_auth",
            "name": "Invalid authentication",
            "ignoreNullFields": True,
            "authentication": {"invalid_key": "value"},
        },
        {
            "In authentication missing required field: 'clientId'",
            "In authentication missing required field: 'clientSecret'",
            "In authentication unused field: 'invalid_key'",
        },
        id="Invalid authentication type",
    )


class TestTransformationYAML:
    @pytest.mark.parametrize("data", list(find_resources("transformation")))
    def test_load_valid_transformation_file(self, data: dict[str, object]) -> None:
        loaded = TransformationYAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        if "authentication" in dumped:
            # Secret is not dumped as per design, so we add it back for comparison
            dumped["authentication"]["clientSecret"] = data["authentication"]["clientSecret"]
        assert dumped == data

    @pytest.mark.parametrize("data", transformation_destination_cases())
    def test_load_valid_transformation_destination_parameters(self, data: dict[str, object]) -> None:
        loaded = TransformationYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_transformation_test_cases()))
    def test_invalid_transformation_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test the validate_resource_yaml function for GroupYAML."""
        warning_list = validate_resource_yaml_pydantic(data, TransformationYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from cognite.client import CogniteClient

from cognite_toolkit._cdf_tk.client.config import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.project import ProjectStatusList
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ParamRequest


class ProjectAPI:
    def __init__(self, config: ToolkitClientConfig, cognite_client: CogniteClient) -> None:
        self._config = config
        self._cognite_client = cognite_client
        self._http_client = HTTPClient(config, split_items_status_codes=set())

    def status(self) -> ProjectStatusList:
        """Retrieve information about the current project."""
        response = self._http_client.request_with_retries(
            ParamRequest(
                endpoint_url=f"{self._config.base_url}/api/v1/projects?withDataModelingStatus=true", method="GET"
            )
        )
        response.raise_for_status()
        body = response.get_first_body()
        return ProjectStatusList._load(body["items"], cognite_client=self._cognite_client)  # type: ignore[arg-type]



from dataclasses import dataclass
from enum import Enum
from functools import lru_cache

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import clean_name


@dataclass(frozen=True)
class FlagMetadata:
    visible: bool
    description: str


class Flags(Enum):
    INTERNAL = FlagMetadata(
        visible=False,
        description="Does nothing",
    )
    IMPORT_CMD = FlagMetadata(
        visible=True,
        description="Enables the import sub application",
    )
    GRAPHQL = FlagMetadata(
        visible=True,
        description="Enables the support for deploying data models as GraphQL schemas",
    )
    MODULE_REPEAT = FlagMetadata(
        visible=True,
        description="Enables the support for repeating modules in the config file",
    )
    AGENTS = FlagMetadata(
        visible=True,
        description="Enables support for Atlas Agents and Agent Tools",
    )
    DUMP_EXTENDED = FlagMetadata(
        visible=True,
        description="Enables support for the dumping Location Filters.",
    )
    PROFILE = FlagMetadata(
        visible=True,
        description="Enables support for the profile command",
    )
    INFIELD = FlagMetadata(
        visible=True,
        description="Enables support for Infield configs",
    )
    DUMP_DATA = FlagMetadata(
        visible=True,
        description="Splits the dump command in dump data and dump config",
    )
    EXIT_ON_WARNING = FlagMetadata(
        visible=True,
        description="Enables the exit on warning feature during the build command",
    )
    MIGRATE = FlagMetadata(
        visible=True,
        description="Enables the migrate command",
    )
    EXTERNAL_LIBRARIES = FlagMetadata(
        visible=True,
        description="Enables the support for external libraries in the config file",
    )
    PURGE_INSTANCES = FlagMetadata(
        visible=True,
        description="Enables the cdf purge instances command",
    )
    DOWNLOAD = FlagMetadata(
        visible=True,
        description="Enables the support for the download command",
    )
    UPLOAD = FlagMetadata(
        visible=True,
        description="Enables the cdf upload command",
    )
    SEARCH_CONFIG = FlagMetadata(
        visible=True,
        description="Enables the support for the search config command",
    )
    FUNCTION_COGNITE_FILE = FlagMetadata(
        visible=True,
        description="Enables support for using Cognite Files to store function code",
    )
    v07 = FlagMetadata(
        visible=False,
        description="Enables features planned for Cognite Toolkit version 0.7.0",
    )
    STREAMS = FlagMetadata(
        visible=True,
        description="Enables the support for the streams resources",
    )

    def is_enabled(self) -> bool:
        return FeatureFlag.is_enabled(self)


class FeatureFlag:
    @staticmethod
    @lru_cache(typed=True)
    def is_enabled(flag: Flags) -> bool:
        return CDFToml.load().alpha_flags.get(clean_name(flag.name), False)



import gzip
from abc import ABC, abstractmethod
from collections.abc import Mapping
from io import TextIOWrapper
from pathlib import Path
from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses


class Compression(ABC):
    encoding = "utf-8-sig"
    newline = "\n"
    name: ClassVar[str]
    file_suffix: ClassVar[str]

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath

    @abstractmethod
    def open(self, mode: Literal["r", "w"]) -> TextIOWrapper:
        """Open the compressed file and return a file-like object."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    @classmethod
    def from_filepath(cls, filepath: Path) -> "Compression":
        return COMPRESSION_BY_SUFFIX.get(filepath.suffix, Uncompressed)(filepath=filepath)

    @classmethod
    def from_name(cls, compression: str) -> "type[Compression]":
        if compression in COMPRESSION_BY_NAME:
            return COMPRESSION_BY_NAME[compression]
        raise ToolkitValueError(
            f"Unknown compression type: {compression}. Available types: {humanize_collection(COMPRESSION_BY_NAME.keys())}."
        )


class Uncompressed(Compression):
    name = "none"
    file_suffix = ""

    def open(self, mode: Literal["r", "w"]) -> TextIOWrapper:
        """Open the file without compression."""
        return self.filepath.open(mode=mode, encoding=self.encoding, newline=self.newline)


class GzipCompression(Compression):
    name = "gzip"
    file_suffix = ".gz"

    def open(self, mode: Literal["r", "w"]) -> TextIOWrapper:
        """Open the gzip compressed file."""
        # MyPy (or gzip) fails to recognize that gzip.open returns a TextIOWrapper
        return gzip.open(self.filepath, mode=f"{mode}t", encoding=self.encoding, newline=self.newline)  # type: ignore[return-value]


_all_compression_types: list[type[Compression]] = get_concrete_subclasses(Compression)  # type: ignore[type-abstract]
COMPRESSION_BY_SUFFIX: Mapping[str, type[Compression]] = {x.file_suffix: x for x in _all_compression_types}
COMPRESSION_BY_NAME: Mapping[str, type[Compression]] = {x.name: x for x in _all_compression_types}



import pytest
from cognite.client.data_classes.data_modeling import NodeApply, NodeApplyList, NodeList, NodeOrEdgeData, Space, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    CreatedSourceSystem,
    InstanceSource,
    ResourceViewMapping,
    ResourceViewMappingApply,
    SpaceSource,
)
from cognite_toolkit._cdf_tk.commands._migrate.data_model import SPACE
from cognite_toolkit._cdf_tk.tk_warnings import IgnoredValueWarning, catch_warnings


@pytest.fixture(scope="session")
def three_sources(toolkit_client: ToolkitClient, toolkit_space: Space) -> NodeList[InstanceSource]:
    nodes = NodeApplyList(
        [
            NodeApply(
                space=toolkit_space.space,
                external_id=f"toolkit_test_migration_{i}",
                sources=[
                    NodeOrEdgeData(
                        InstanceSource.get_source(),
                        {
                            "resourceType": "asset",
                            "id": i,
                            "preferredConsumerViewId": {
                                "space": "cdf_cdm",
                                "externalId": "CogniteAsset",
                                "version": "v1",
                            }
                            if i < 2
                            else {"invalid": "json"},
                        },
                    )
                ],
            )
            for i in range(3)
        ]
    )

    _ = toolkit_client.data_modeling.instances.apply(nodes)

    created = toolkit_client.data_modeling.instances.retrieve_nodes(nodes.as_ids(), node_cls=InstanceSource)
    assert len(created) == 3, "Expected 3 mappings to be created"
    return created


class TestInstanceSourceAPI:
    def test_retrieve_mappings(self, toolkit_client: ToolkitClient, three_sources: NodeList[InstanceSource]) -> None:
        ids = [instance_source.as_asset_centric_id() for instance_source in three_sources]

        with catch_warnings(IgnoredValueWarning) as warnings:
            retrieved = toolkit_client.migration.instance_source.retrieve(ids)

        assert retrieved.dump() == three_sources.dump(), "Failed to retrieve instance source using asset-centric IDs"
        assert len(warnings) == 1, "Expected one warning about invalid JSON in preferredConsumerViewId"
        has_preferred_consumer_view = [
            item.as_id() for item in retrieved if item.preferred_consumer_view_id is not None
        ]
        assert has_preferred_consumer_view == [item.as_id() for item in three_sources[:2]]
        no_preferred_consumer_view = [item.as_id() for item in retrieved if item.preferred_consumer_view_id is None]
        assert no_preferred_consumer_view == [three_sources[2].as_id()], (
            "Expected last item to have no preferred consumer view"
        )

    def test_list_mappings(self, toolkit_client: ToolkitClient, three_sources: NodeList[InstanceSource]) -> None:
        assert all(source.resource_type == "asset" for source in three_sources)
        listed = toolkit_client.migration.instance_source.list("asset", limit=10)
        assert len(listed) >= 3, "Expected at least 3 instance sources to be listed"


class TestResourceViewMappingAPI:
    def test_create_retrieve_list_delete(self, toolkit_client: ToolkitClient) -> None:
        source = ResourceViewMappingApply(
            external_id="test_view_source",
            resource_type="asset",
            view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            property_mapping={
                "name": "name",
                "description": "description",
            },
        )

        created: ResourceViewMapping | None = None
        try:
            created = toolkit_client.migration.resource_view_mapping.upsert(source)

            assert created.external_id == source.external_id

            retrieved = toolkit_client.migration.resource_view_mapping.retrieve(source.external_id)

            assert retrieved.external_id == source.external_id

            listed = toolkit_client.migration.resource_view_mapping.list(resource_type="asset")
            assert len(listed) > 0
            existing = {vs.external_id for vs in listed}
            assert source.external_id in existing, "Expected the created view source to be listed"

            deleted = toolkit_client.migration.resource_view_mapping.delete(source.external_id)

            assert deleted == created.as_id(), "Expected the deleted view source to match the created one"
        finally:
            if created:
                # Clean up by deleting the created view source if it exists
                toolkit_client.data_modeling.instances.delete(source.as_id())


@pytest.fixture(scope="session")
def created_source_system(toolkit_client: ToolkitClient, toolkit_space: Space) -> str:
    source = "toolkit_test_created_source_system"
    node = NodeApply(
        space=toolkit_space.space,
        external_id=source,
        sources=[
            NodeOrEdgeData(
                CreatedSourceSystem.get_source(),
                {"source": source},
            )
        ],
    )
    if not toolkit_client.data_modeling.instances.retrieve_nodes([node.as_id()], node_cls=CreatedSourceSystem):
        _ = toolkit_client.data_modeling.instances.apply(node)
    return source


class TestCreatedSourceSystemAPI:
    @pytest.mark.usefixtures("created_source_system")
    def test_list(self, toolkit_client: ToolkitClient) -> None:
        created = toolkit_client.migration.created_source_system.list(limit=1)
        assert len(created) == 1

    def test_retrieve(self, created_source_system: str, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.migration.created_source_system.retrieve([created_source_system])
        assert len(retrieved) == 1
        node = retrieved[0]
        assert node.external_id == created_source_system
        assert node.source == created_source_system


@pytest.fixture(scope="session")
def space_source(toolkit_client) -> SpaceSource:
    node = NodeApply(
        space=SPACE.space,
        external_id="toolkit_test_space_source_instance",
        sources=[
            NodeOrEdgeData(
                SpaceSource.get_source(),
                {
                    "instanceSpace": "the_instance_space",
                    "dataSetId": 12345,
                    "dataSetExternalId": "the_classic_data_set",
                },
            )
        ],
    )
    if not toolkit_client.data_modeling.instances.retrieve_nodes([node.as_id()], node_cls=SpaceSource):
        _ = toolkit_client.data_modeling.instances.apply(node)
    retrieved = toolkit_client.data_modeling.instances.retrieve_nodes([node.as_id()], node_cls=SpaceSource)
    return retrieved[0]


class TestSpaceSourceAPI:
    def test_retrieve_data_set_id(self, space_source: SpaceSource, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.migration.space_source.retrieve(data_set_id=space_source.data_set_id)
        assert retrieved is not None
        assert retrieved.dump() == space_source.dump(), "Failed to retrieve space source by data set ID"

    def test_retrieve_data_set_external_id(self, space_source: SpaceSource, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.migration.space_source.retrieve(
            data_set_external_id=space_source.data_set_external_id
        )

        assert retrieved is not None
        assert retrieved.dump() == space_source.dump(), "Failed to retrieve space source by classic external ID"

    def test_retrieve_not_found(self, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.migration.space_source.retrieve(data_set_id=999999999)
        assert retrieved is None, "Expected None when retrieving non-existing space source"

    def test_retrieve_data_external_id_not_found(self, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.migration.space_source.retrieve(data_set_external_id="non_existing_external_id")
        assert retrieved is None, "Expected None when retrieving non-existing space source by external ID"

    @pytest.mark.usefixtures("space_source")
    def test_list(self, toolkit_client: ToolkitClient) -> None:
        listed = toolkit_client.migration.space_source.list(limit=1)
        assert len(listed) == 1



from .canvas import MigrationCanvasCommand
from .command import MigrationCommand
from .prepare import MigrationPrepareCommand

__all__ = [
    "MigrationCanvasCommand",
    "MigrationCommand",
    "MigrationPrepareCommand",
]



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.hosted_extractor_source import HostedExtractorSourceYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_hosted_extractor_source_test_cases() -> Iterable:
    # Invalid type
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "invalid",
            "host": "http://example.com",
            "published": True,
            "keyName": "apiKey",
            "keyValue": "secret",
        },
        {
            "Invalid hosted extractor source type='invalid'. Expected one of eventhub, kafka, mqtt3, mqtt5 or rest",
        },
        id="Invalid source type",
    )
    # Invalid eventhub (missing required event_hub_name)
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "eventhub",
            "host": "host",
            "keyName": "key",
            "keyValue": "secret",
            # missing eventHubName
        },
        {
            "Missing required field: 'eventHubName'",
        },
        id="EventHubSource missing event_hub_name",
    )
    # Invalid kafka (missing bootstrap_brokers)
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "kafka",
            # missing bootstrapBrokers
        },
        {
            "Missing required field: 'bootstrapBrokers'",
        },
        id="KafkaSource missing bootstrap_brokers",
    )
    # Invalid rest (missing host)
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "rest",
            # missing host
        },
        {
            "Missing required field: 'host'",
        },
        id="RESTSource missing host",
    )
    # Invalid mqtt3 (missing host)
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "mqtt3",
            # missing host
        },
        {
            "Missing required field: 'host'",
        },
        id="MQTT3Source missing host",
    )
    # Invalid mqtt5 (missing host)
    yield pytest.param(
        {
            "externalId": "mySource",
            "type": "mqtt5",
            # missing host
        },
        {
            "Missing required field: 'host'",
        },
        id="MQTT5Source missing host",
    )
    # RESTSource with BasicAuthentication (missing password)
    yield pytest.param(
        {
            "externalId": "restBasicMissingPassword",
            "type": "rest",
            "host": "api.example.com",
            "authentication": {
                "type": "basic",
                "username": "user",
                # missing password
            },
        },
        {"In authentication missing required field: 'password'"},
        id="RESTSource BasicAuthentication missing password",
    )
    # RESTSource with ClientCredentials (missing client_secret)
    yield pytest.param(
        {
            "externalId": "restClientCredMissingSecret",
            "type": "rest",
            "host": "api.example.com",
            "authentication": {
                "type": "clientCredentials",
                "client_id": "id",
                "token_url": "https://token.url",
                "scope": "scope",
                # missing client_secret
            },
        },
        {
            "In authentication missing required field: 'clientId'",
            "In authentication missing required field: 'clientSecret'",
            "In authentication missing required field: 'tokenUrl'",
            "In authentication unused field: 'client_id'",
            "In authentication unused field: 'token_url'",
        },
        id="RESTSource ClientCredentials missing client_secret",
    )
    # RESTSource with QueryCredentials (missing value)
    yield pytest.param(
        {
            "externalId": "restQueryCredMissingValue",
            "type": "rest",
            "host": "api.example.com",
            "authentication": {
                "type": "query",
                "key": "api_key",
                # missing value
            },
        },
        {"In authentication missing required field: 'value'"},
        id="RESTSource QueryCredentials missing value",
    )
    # RESTSource with HeaderCredentials (missing value)
    yield pytest.param(
        {
            "externalId": "restHeaderCredMissingValue",
            "type": "rest",
            "host": "api.example.com",
            "authentication": {
                "type": "header",
                "key": "Authorization",
                # missing value
            },
        },
        {"In authentication missing required field: 'value'"},
        id="RESTSource HeaderCredentials missing value",
    )
    # RESTSource with ScramSha256 (invalid type)
    # RESTSource with ScramSha256 (invalid type)
    yield pytest.param(
        {
            "externalId": "restSourceWithInvalidAuth",
            "type": "rest",
            "host": "api.example.com",
            "authentication": {
                "type": "scramSha256",
                "username": "user",
                "password": "secret",
            },
        },
        {
            "In field authentication invalid authentication type 'scramSha256' for REST source. Expected one of basic, clientCredentials, header or query"
        },
        id="RESTSource with invalid auth type",
    )
    # KafkaSource with ScramSha256 (missing password)
    yield pytest.param(
        {
            "externalId": "kafkaScram256MissingPassword",
            "type": "kafka",
            "bootstrapBrokers": [{"host": "broker", "port": 9092}],
            "authentication": {
                "type": "scramSha256",
                "username": "user",
                # missing password
            },
        },
        {"In authentication missing required field: 'password'"},
        id="KafkaSource ScramSha256 missing password",
    )
    # KafkaSource with BasicAuthentication (missing password)
    yield pytest.param(
        {
            "externalId": "kafkaBasicMissingPassword",
            "type": "kafka",
            "bootstrapBrokers": [{"host": "broker", "port": 9092}],
            "authentication": {
                "type": "basic",
                "username": "user",
                # missing password
            },
        },
        {"In authentication missing required field: 'password'"},
        id="KafkaSource BasicAuthentication missing password",
    )
    # KafkaSource with ClientCredentials (missing client_secret)
    yield pytest.param(
        {
            "externalId": "kafkaClientCredMissingSecret",
            "type": "kafka",
            "bootstrapBrokers": [{"host": "broker", "port": 9092}],
            "authentication": {
                "type": "clientCredentials",
                "clientId": "id",
                "tokenUrl": "https://token.url",
                "scope": "scope",
                # missing client_secret
            },
        },
        {"In authentication missing required field: 'clientSecret'"},
        id="KafkaSource ClientCredentials missing client_secret",
    )
    # KafkaSource with QueryCredentials (invalid type)
    yield pytest.param(
        {
            "externalId": "kafkaQueryCredMissingValue",
            "type": "kafka",
            "bootstrapBrokers": [{"host": "broker", "port": 9092}],
            "authentication": {
                "type": "query",
                "key": "api_key",
                "value": "api_secret",
            },
        },
        {
            "In field authentication invalid authentication type 'query' for Kafka source. Expected one of basic, clientCredentials, scramSha256 or scramSha512"
        },
        id="KafkaSource QueryCredentials invalid type for Kafka",
    )


class TestHostedExtractorSourceYAML:
    @pytest.mark.parametrize("data", list(find_resources("Source", resource_dir="hosted_extractors")))
    def test_load_valid_hosted_extractor_source(self, data: dict[str, object]) -> None:
        loaded = HostedExtractorSourceYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True, mode="json") == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_hosted_extractor_source_test_cases()))
    def test_invalid_hosted_extractor_source_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, HostedExtractorSourceYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from .agent import AgentCRUD
from .auth import GroupAllScopedCRUD, GroupCRUD, SecurityCategoryCRUD
from .classic import AssetCRUD, EventCRUD, SequenceCRUD, SequenceRowCRUD
from .configuration import SearchConfigCRUD
from .data_organization import DataSetsCRUD, LabelCRUD
from .datamodel import (
    ContainerCRUD,
    DataModelCRUD,
    EdgeCRUD,
    GraphQLCRUD,
    NodeCRUD,
    SpaceCRUD,
    ViewCRUD,
)
from .extraction_pipeline import ExtractionPipelineConfigCRUD, ExtractionPipelineCRUD
from .fieldops import InFieldLocationConfigCRUD, InfieldV1CRUD
from .file import CogniteFileCRUD, FileMetadataCRUD
from .function import FunctionCRUD, FunctionScheduleCRUD
from .group_scoped import GroupResourceScopedCRUD
from .hosted_extractors import (
    HostedExtractorDestinationCRUD,
    HostedExtractorJobCRUD,
    HostedExtractorMappingCRUD,
    HostedExtractorSourceCRUD,
)
from .industrial_tool import StreamlitCRUD
from .location import LocationFilterCRUD
from .migration import ResourceViewMappingCRUD
from .raw import RawDatabaseCRUD, RawTableCRUD
from .relationship import RelationshipCRUD
from .robotics import (
    RobotCapabilityCRUD,
    RoboticFrameCRUD,
    RoboticLocationCRUD,
    RoboticMapCRUD,
    RoboticsDataPostProcessingCRUD,
)
from .three_d_model import ThreeDModelCRUD
from .timeseries import DatapointSubscriptionCRUD, TimeSeriesCRUD
from .transformation import TransformationCRUD, TransformationNotificationCRUD, TransformationScheduleCRUD
from .workflow import WorkflowCRUD, WorkflowTriggerCRUD, WorkflowVersionCRUD

__all__ = [
    "AgentCRUD",
    "AssetCRUD",
    "CogniteFileCRUD",
    "ContainerCRUD",
    "DataModelCRUD",
    "DataSetsCRUD",
    "DatapointSubscriptionCRUD",
    "EdgeCRUD",
    "EventCRUD",
    "ExtractionPipelineCRUD",
    "ExtractionPipelineConfigCRUD",
    "FileMetadataCRUD",
    "FunctionCRUD",
    "FunctionScheduleCRUD",
    "GraphQLCRUD",
    "GroupAllScopedCRUD",
    "GroupCRUD",
    "GroupResourceScopedCRUD",
    "HostedExtractorDestinationCRUD",
    "HostedExtractorJobCRUD",
    "HostedExtractorMappingCRUD",
    "HostedExtractorSourceCRUD",
    "InFieldLocationConfigCRUD",
    "InfieldV1CRUD",
    "LabelCRUD",
    "LocationFilterCRUD",
    "NodeCRUD",
    "RawDatabaseCRUD",
    "RawTableCRUD",
    "RelationshipCRUD",
    "ResourceViewMappingCRUD",
    "RobotCapabilityCRUD",
    "RoboticFrameCRUD",
    "RoboticLocationCRUD",
    "RoboticMapCRUD",
    "RoboticsDataPostProcessingCRUD",
    "SearchConfigCRUD",
    "SecurityCategoryCRUD",
    "SequenceCRUD",
    "SequenceRowCRUD",
    "SpaceCRUD",
    "StreamlitCRUD",
    "ThreeDModelCRUD",
    "TimeSeriesCRUD",
    "TransformationCRUD",
    "TransformationNotificationCRUD",
    "TransformationScheduleCRUD",
    "ViewCRUD",
    "WorkflowCRUD",
    "WorkflowTriggerCRUD",
    "WorkflowVersionCRUD",
]



import sys
from dataclasses import dataclass, field, fields
from functools import lru_cache
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.data_modeling import NodeId, ViewId
from cognite.client.utils._auxiliary import to_camel_case

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class ChartObject(CogniteObject):
    # ChartObjects are used in the frontend and the backend does not do any validation of these fields.
    # Therefore, to ensure that we do not lose any data, we store unknown fields in a separate dictionary.
    # This allows unknown fields to be preserved when loading and dumping ChartObjects
    # (serialization and deserialization).
    _unknown_fields: dict[str, object] | None = field(default=None, init=False, repr=False)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a ChartObject from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        instance._unknown_fields = {k: v for k, v in resource.items() if k not in cls._known_camel_case_props()}
        return instance

    @classmethod
    @lru_cache(maxsize=1)
    def _known_camel_case_props(cls) -> set[str]:
        return {to_camel_case(f.name) for f in fields(cls)}

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the ChartObject to a dictionary."""
        data = super().dump(camel_case=camel_case)
        if self._unknown_fields:
            data.update(self._unknown_fields)
        return data


@dataclass
class UserInfo(ChartObject):
    id: str | None = None
    email: str | None = None
    display_name: str | None = None


@dataclass
class ChartSettings(ChartObject):
    show_y_axis: bool = True
    show_min_max: bool = True
    show_gridlines: bool = True
    merge_units: bool = False


@dataclass
class ThresholdFilter(ChartObject):
    min_unit: str | None = None
    max_unit: str | None = None


@dataclass
class ChartCall(ChartObject):
    id: str | None = None
    hash: int | None = None
    call_id: str | None = None
    call_date: int | None = None
    status: str | None = None


@dataclass
class SubSetting(ChartObject):
    auto_align: bool | None = None


@dataclass
class FlowElement(ChartObject):
    id: str | None = None
    type: str | None = None
    position: tuple[float | None, float | None] | None = None
    data: dict[str, object] | None = None


@dataclass
class Flow(ChartObject):
    zoom: float | None = None
    elements: list[FlowElement] | None = None
    position: tuple[float | None, float | None] | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        if self.elements:
            data["elements"] = [el.dump(camel_case=camel_case) for el in self.elements]
        return data

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a Flow object from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        if "elements" in resource:
            instance.elements = [FlowElement._load(el, cognite_client=cognite_client) for el in resource["elements"]]
        return instance


@dataclass
class ChartSource(ChartObject):
    type: str | None = None
    id: str | None = None


@dataclass
class BaseChartElement(ChartObject):
    type: str | None = None
    id: str | None = None
    name: str | None = None
    color: str | None = None
    enabled: bool | None = None
    line_weight: float | None = None
    line_style: str | None = None
    interpolation: str | None = None
    unit: str | None = None
    preferred_unit: str | None = None
    created_at: int | None = None
    range: tuple[float | None, float | None] | None = None
    description: str | None = None


@dataclass
class ChartCoreTimeseries(BaseChartElement):
    node_reference: NodeId | None = None
    view_reference: ViewId | None = None
    display_mode: str | None = None


@dataclass
class ChartTimeseries(BaseChartElement):
    ts_id: int | None = None
    ts_external_id: str | None = None
    display_mode: str | None = None
    original_unit: str | None = None


@dataclass
class ChartWorkflow(BaseChartElement):
    version: str | None = None
    settings: SubSetting | None = None
    flow: Flow | None = None
    calls: list[ChartCall] | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        if self.settings:
            data["settings"] = self.settings.dump(camel_case=camel_case)
        if self.flow:
            data["flow"] = self.flow.dump(camel_case=camel_case)
        if self.calls:
            data["calls"] = [c.dump(camel_case=camel_case) for c in self.calls]
        return data

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a ChartWorkflow object from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        if "settings" in resource:
            instance.settings = SubSetting._load(resource["settings"], cognite_client=cognite_client)
        if "flow" in resource:
            instance.flow = Flow._load(resource["flow"], cognite_client=cognite_client)
        if "calls" in resource:
            instance.calls = [ChartCall._load(call, cognite_client=cognite_client) for call in resource["calls"]]
        return instance


@dataclass
class ChartThreshold(BaseChartElement):
    visible: bool | None = None
    source_id: str | None = None
    upper_limit: float | None = None
    filter: ThresholdFilter | None = None
    calls: list[ChartCall] | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        if self.filter:
            data["filter"] = self.filter.dump(camel_case=camel_case)
        if self.calls:
            data["calls"] = [c.dump(camel_case=camel_case) for c in self.calls]
        return data

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a ChartThreshold object from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        if "filter" in resource:
            instance.filter = ThresholdFilter._load(resource["filter"], cognite_client=cognite_client)
        if "calls" in resource:
            instance.calls = [ChartCall._load(call, cognite_client=cognite_client) for call in resource["calls"]]
        return instance


@dataclass
class ChartScheduledCalculation(BaseChartElement):
    version: str | None = None
    settings: SubSetting | None = None
    flow: Flow | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        if self.settings:
            data["settings"] = self.settings.dump(camel_case=camel_case)
        if self.flow:
            data["flow"] = self.flow.dump(camel_case=camel_case)
        return data

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a ChartScheduledCalculation object from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        if "settings" in resource:
            instance.settings = SubSetting._load(resource["settings"], cognite_client=cognite_client)
        if "flow" in resource:
            instance.flow = Flow._load(resource["flow"], cognite_client=cognite_client)
        return instance


@dataclass
class ChartData(ChartObject):
    version: int | None = None
    name: str | None = None
    date_from: str | None = None
    date_to: str | None = None
    user_info: UserInfo | None = None
    live_mode: bool | None = None
    time_series_collection: list[ChartTimeseries] | None = None
    core_timeseries_collection: list[ChartCoreTimeseries] | None = None
    workflow_collection: list[ChartWorkflow] | None = None
    source_collection: list[ChartSource] | None = None
    threshold_collection: list[ChartThreshold] | None = None
    scheduled_calculation_collection: list[ChartScheduledCalculation] | None = None
    settings: ChartSettings | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the ChartData object to a dictionary."""
        data = super().dump(camel_case=camel_case)
        list_attrs = [
            "time_series_collection",
            "core_timeseries_collection",
            "workflow_collection",
            "source_collection",
            "threshold_collection",
            "scheduled_calculation_collection",
        ]
        for attr_name in list_attrs:
            if collection := getattr(self, attr_name):
                key = to_camel_case(attr_name) if camel_case else attr_name
                data[key] = [item.dump(camel_case=camel_case) for item in collection]

        single_attrs_map = {
            "user_info": "userInfo",
            "settings": "settings",
        }
        for attr_name, camel_key in single_attrs_map.items():
            if item := getattr(self, attr_name):
                key = camel_key if camel_case else attr_name
                data[key] = item.dump(camel_case=camel_case)
        return data

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a ChartData object from a dictionary."""
        instance = super()._load(resource, cognite_client=cognite_client)
        collections_map = [
            ("timeSeriesCollection", "time_series_collection", ChartTimeseries),
            ("coreTimeseriesCollection", "core_timeseries_collection", ChartCoreTimeseries),
            ("workflowCollection", "workflow_collection", ChartWorkflow),
            ("sourceCollection", "source_collection", ChartSource),
            ("thresholdCollection", "threshold_collection", ChartThreshold),
            ("scheduledCalculationCollection", "scheduled_calculation_collection", ChartScheduledCalculation),
        ]
        for resource_key, attr_name, subclass in collections_map:
            if resource_key in resource:
                setattr(
                    instance,
                    attr_name,
                    [subclass._load(item, cognite_client=cognite_client) for item in resource[resource_key]],  # type: ignore[attr-defined]
                )
        attribute_map = [
            ("userInfo", "user_info", UserInfo),
            ("settings", "settings", ChartSettings),
        ]
        for resource_key, attr_name, subclass in attribute_map:
            if resource_key in resource:
                setattr(
                    instance,
                    attr_name,
                    subclass._load(resource[resource_key], cognite_client=cognite_client),  # type: ignore[attr-defined]
                )

        return instance



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import DataModelYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestDataModelYAML:
    """Test suite for DataModelYAML class."""

    @pytest.mark.parametrize("data", list(find_resources("DataModel")))
    def test_load_valid_data_model_from_resources(self, data: dict[str, object]) -> None:
        """Test loading valid data models from resource files."""
        loaded = DataModelYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                },
                id="minimal-no-views",
            ),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                    "views": [
                        {
                            "space": "view_space1",
                            "externalId": "view1",
                            "version": "v1",
                            "type": "view",
                        },
                        {
                            "space": "view_space2",
                            "externalId": "view2",
                            "version": "v2",
                            "type": "view",
                        },
                    ],
                },
                id="with-multiple-views",
            ),
        ],
    )
    def test_valid_data_models(self, data: dict) -> None:
        """Test various valid data model configurations."""
        loaded = DataModelYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data,expected_errors",
        [
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    # Missing 'version'
                },
                ["Missing required field: 'version'"],
                id="missing-required-field",
            ),
            pytest.param(
                {
                    "space": "",  # Empty space
                    "externalId": "my_model",
                    "version": "v1",
                },
                ["In field space string should have at least 1 character"],
                id="invalid-space",
            ),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "a" * 256,  # External ID too long (max 255)
                    "version": "v1",
                },
                ["In field externalId string should have at most 255 characters"],
                id="invalid-external-id",
            ),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "a" * 44,  # Version too long (max 43)
                },
                ["In field version string should have at most 43 characters"],
                id="invalid-version",
            ),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                    "name": "a" * 256,  # Name too long (max 255)
                },
                ["In field name string should have at most 255 characters"],
                id="invalid-name",
            ),
        ],
    )
    def test_invalid_data_models(self, data: dict, expected_errors: list[str]) -> None:
        """Test that invalid data models are properly rejected with appropriate error messages."""
        warning_list = validate_resource_yaml_pydantic(data, DataModelYAML, Path("test.yaml"))
        assert len(warning_list) >= 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert list(format_warning.errors) == expected_errors

    @pytest.mark.parametrize(
        "data,expected_errors",
        [
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                    "views": [
                        {
                            "space": "view_space",
                            "externalId": "my_view",
                        }
                    ],
                },
                ["In views[1] missing required field: 'version'"],
                id="view-missing-version",
            ),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                    "views": [
                        {
                            "space": "view_space",
                            "externalId": "my_view",
                            "version": "v1",
                            "type": "invalid_type",
                        }
                    ],
                },
                ["In views[1].type input should be 'view'. Got 'invalid_type'."],
                id="view-invalid-type",
            ),
        ],
    )
    def test_invalid_views_in_data_model(self, data: dict, expected_errors: list[str]) -> None:
        """Test that data models with invalid views are properly rejected."""
        warning_list = validate_resource_yaml_pydantic(data, DataModelYAML, Path("test.yaml"))
        assert len(warning_list) >= 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert list(format_warning.errors) == expected_errors

    @pytest.mark.parametrize(
        "data,expected_errors",
        [
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "my_model",
                    "version": "v1",
                    "unknown_field": "some_value",
                },
                ["Unused field: 'unknown_field'"],
                id="single-unknown-field",
            ),
        ],
    )
    def test_data_model_with_extra_fields(self, data: dict, expected_errors: str) -> None:
        """Test that data models with extra/unused properties are properly rejected."""
        warning_list = validate_resource_yaml_pydantic(data, DataModelYAML, Path("test.yaml"))
        assert len(warning_list) >= 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert list(format_warning.errors) == expected_errors



from collections.abc import Iterable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.mark import ParameterSet
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes import FileMetadata, FileMetadataWrite, FileMetadataWriteList

from cognite_toolkit._cdf_tk.cruds import FileMetadataCRUD
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.approval_client.client import LookUpAPIMock


def file_metadata_config_cases() -> Iterable[ParameterSet]:
    yield pytest.param(
        """externalId: sharepointABC
name: A file.txt
dataSetExternalId: ds_files
source: sharepointABC
""",
        ["1.A file.txt"],
        FileMetadataWriteList(
            [
                FileMetadataWrite(
                    external_id="sharepointABC",
                    source="sharepointABC",
                    name="A file.txt",
                    data_set_id=LookUpAPIMock.create_id("ds_files"),
                )
            ]
        ),
        id="Single file as mapping",
    )
    yield pytest.param(
        """- externalId: sharepointABC
  name: A file.txt
  dataSetExternalId: ds_files
  source: sharepointABC
- externalId: sharepointABC2
  name: Another file.txt
  dataSetExternalId: ds_files
  source: sharepointABC
""",
        ["1.A file.txt", "1.Another file.txt"],
        FileMetadataWriteList(
            [
                FileMetadataWrite(
                    external_id="sharepointABC",
                    source="sharepointABC",
                    name="A file.txt",
                    data_set_id=LookUpAPIMock.create_id("ds_files"),
                ),
                FileMetadataWrite(
                    external_id="sharepointABC2",
                    source="sharepointABC",
                    name="Another file.txt",
                    data_set_id=LookUpAPIMock.create_id("ds_files"),
                ),
            ]
        ),
        id="Multiple files as array",
    )


class TestLoadResources:
    @pytest.mark.parametrize("yaml_content, files, expected", list(file_metadata_config_cases()))
    def test_load_resources(
        self,
        yaml_content: str,
        files: list[str],
        expected: FileMetadataWriteList,
        toolkit_client_approval: ApprovalToolkitClient,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = FileMetadataCRUD(toolkit_client_approval.mock_client, None)
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = yaml_content
        filepath.parent.glob.return_value = [Path(f) for f in files]
        raw_list = loader.load_resource_file(filepath, {})
        resources = FileMetadataWriteList([loader.load_resource(item, is_dry_run=False) for item in raw_list])

        assert resources.dump() == expected.dump()

    @staticmethod
    def always_existing_path(*_) -> Path:
        path = MagicMock(spec=Path)
        path.exists.return_value = True
        return path

    def test_dump_file_metadata_without_dataset(
        self, monkeypatch: MonkeyPatch, toolkit_client_approval: ApprovalToolkitClient
    ) -> None:
        metadata = FileMetadata("my_file", name="my_file.txt", mime_type="text/plain")
        loader = FileMetadataCRUD.create_loader(toolkit_client_approval.mock_client)

        dumped = loader.dump_resource(metadata)

        assert dumped == {
            "externalId": "my_file",
            "name": "my_file.txt",
            "mimeType": "text/plain",
        }



from collections.abc import Iterable
from pathlib import Path
from typing import get_args

import pytest

from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.resource_classes.agent import AgentYAML
from cognite_toolkit._cdf_tk.resource_classes.agent_tools import (
    AgentInstanceSpaces,
    AgentInstanceSpacesDefinition,
    AgentTool,
    AgentToolDefinition,
)
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.data import COMPLETE_ORG_ALPHA_FLAGS
from tests.test_unit.utils import find_resources


def invalid_test_cases() -> Iterable:
    yield pytest.param(
        {
            "externalId": "my_space",
        },
        {"Missing required field: 'name'"},
        id="missing-required-field",
    )
    yield pytest.param(
        {
            "externalId": "",  # Empty string - violates min_length=1
            "name": "",  # Empty string - violates min_length=1
            "description": "a" * 1025,  # Too long - violates max_length=1024
            "model": "invalid-model",  # Not in Model literal enum
            "tools": [{"type": "invalid"}] * 21,  # Too many tools - violates max_length=20
        },
        {
            "In field description string should have at most 1024 characters",
            "In field externalId string should have at least 1 character",
            "In field model input should be 'azure/o3', 'azure/o4-mini', 'azure/gpt-4o', "
            "'azure/gpt-4o-mini', 'azure/gpt-4.1', 'azure/gpt-4.1-nano', "
            "'azure/gpt-4.1-mini', 'azure/gpt-5', 'azure/gpt-5-mini', 'azure/gpt-5-nano', "
            "'gcp/gemini-2.5-pro', 'gcp/gemini-2.5-flash', 'aws/claude-4-sonnet', "
            "'aws/claude-4-opus', 'aws/claude-4.1-opus' or 'aws/claude-3.5-sonnet'. Got "
            "'invalid-model'.",
            "In field name string should have at least 1 character",
            "In field tools list should have at most 20 items after validation, not 21",
        },
        id="multiple-validation-errors",
    )
    yield pytest.param(
        {
            "externalId": "a" * 129,  # Too long - violates max_length=128
            "name": "a" * 256,  # Too long - violates max_length=255
            "instructions": "a" * 32001,  # Too long - violates max_length=32000
            "tools": [
                {
                    "type": "askDocument",
                    "name": "",  # Empty name - violates min_length=1
                    "description": "short",  # Too short - violates min_length=10
                }
            ],
        },
        {
            "In field externalId string should have at most 128 characters",
            "In field instructions string should have at most 32000 characters",
            "In field name string should have at most 255 characters",
            "In tools[1].askDocument.description string should have at least 10 characters",
            "In tools[1].askDocument.name string should have at least 1 character",
        },
        id="length-validation-errors",
    )
    yield pytest.param(
        {
            "externalId": "valid_id",
            "name": "Valid Name",
            "tools": [
                {
                    "type": "queryKnowledgeGraph",
                    "name": "test_tool",
                    "description": "This is a valid description for the tool",
                    "configuration": {
                        "dataModels": [],  # Empty list - violates min_length=1
                        "instanceSpaces": {
                            "type": "manual",
                            # Missing required 'spaces' field for manual type
                        },
                        "version": "v3",  # Invalid literal - not in ["v1", "v2"]
                    },
                }
            ],
        },
        {
            "In tools[1].queryKnowledgeGraph.configuration.dataModels list should have at "
            "least 1 item after validation, not 0",
            "In tools[1].queryKnowledgeGraph.configuration.instanceSpaces.manual missing required field: 'spaces'",
            "In tools[1].queryKnowledgeGraph.configuration.version input should be 'v1' or 'v2'. Got 'v3'.",
        },
        id="nested-tool-validation-errors",
    )
    yield pytest.param(
        {
            "externalId": "\x00invalid",  # Contains null character - violates pattern
            "name": "Valid Name",
            "tools": [
                {
                    "type": "queryKnowledgeGraph",
                    "name": "\x00tool",  # Contains null character - violates pattern
                    "description": "a" * 1025,  # Too long - violates max_length=1024
                    "configuration": {
                        "dataModels": [
                            {
                                "space": "invalid space!",  # Invalid pattern - contains space and special char
                                "external_id": "123invalid",  # Invalid pattern - starts with number
                                "version": "v@1",  # Invalid pattern - contains special char
                                "viewExternalIds": ["valid_view"] * 11,  # Too many items - violates max_length=10
                            }
                        ]
                        * 81,  # Too many data models - violates max_length=80
                        "instanceSpaces": {"type": "all"},
                    },
                }
            ],
        },
        {
            "In field externalId string should match pattern '^[^\\x00]{1,128}$'",
            "In tools[1].queryKnowledgeGraph.configuration.dataModels list should have at "
            "most 80 items after validation, not 81",
            "In tools[1].queryKnowledgeGraph.description string should have at most 1024 characters",
            "In tools[1].queryKnowledgeGraph.name string should match pattern '^[^\\x00]{1,64}$'",
        },
        id="pattern-and-nested-validation-errors",
    )
    yield pytest.param(
        {
            "external_id": 123,  # Wrong type - should be string
            "name": None,  # Wrong type - should be string
            "description": 456,  # Wrong type - should be string or None
            "instructions": [],  # Wrong type - should be string or None
            "model": True,  # Wrong type - should be Model literal
            "tools": "not_a_list",  # Wrong type - should be list
        },
        {
            "In field description input should be a valid string. Got 456 of type int. "
            "Hint: Use double quotes to force string.",
            "In field instructions input should be a valid string. Got [] of type list. "
            "Hint: Use double quotes to force string.",
            "In field model input should be 'azure/o3', 'azure/o4-mini', 'azure/gpt-4o', "
            "'azure/gpt-4o-mini', 'azure/gpt-4.1', 'azure/gpt-4.1-nano', "
            "'azure/gpt-4.1-mini', 'azure/gpt-5', 'azure/gpt-5-mini', 'azure/gpt-5-nano', "
            "'gcp/gemini-2.5-pro', 'gcp/gemini-2.5-flash', 'aws/claude-4-sonnet', "
            "'aws/claude-4-opus', 'aws/claude-4.1-opus' or 'aws/claude-3.5-sonnet'. Got "
            "True.",
            "In field name input should be a valid string. Got None of type NoneType. "
            "Hint: Use double quotes to force string.",
            "In field tools input should be a valid list. Got 'not_a_list'.",
            "Missing required field: 'externalId'",
            "Unused field: 'external_id'",
        },
        id="type-validation-errors",
    )


class TestAgentYAML:
    @pytest.mark.parametrize("data", list(find_resources("Agent", base=COMPLETE_ORG_ALPHA_FLAGS / MODULES)))
    def test_load_valid(self, data: dict[str, object]) -> None:
        loaded = AgentYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_test_cases()))
    def test_invalid_model_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, AgentYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors

    def test_tools_are_in_union(self) -> None:
        all_agent_tools = get_concrete_subclasses(AgentToolDefinition)
        all_union_agent_tools = get_args(AgentTool.__args__[0])
        missing = set(all_agent_tools) - set(all_union_agent_tools)
        assert not missing, (
            f"The following AgentTools subclasses are "
            f"missing from the AgentTool union: {humanize_collection([cls.__name__ for cls in missing])}"
        )

    def test_instance_spaces_are_in_union(self) -> None:
        all_instance_spaces = get_concrete_subclasses(AgentInstanceSpacesDefinition)
        all_union_instance_spaces = get_args(AgentInstanceSpaces.__args__[0])
        missing = set(all_instance_spaces) - set(all_union_instance_spaces)
        assert not missing, (
            f"The following AgentInstanceSpaces subclasses are "
            f"missing from the AgentInstanceSpaces union: {humanize_collection([cls.__name__ for cls in missing])}"
        )



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, ClassVar, cast

from cognite.client._proto.data_point_insertion_request_pb2 import DataPointInsertionItem, DataPointInsertionRequest
from cognite.client._proto.data_point_list_response_pb2 import DataPointListResponse
from cognite.client._proto.data_points_pb2 import (
    NumericDatapoint,
    NumericDatapoints,
    StringDatapoint,
    StringDatapoints,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    _EpochConverter,
    _Float64Converter,
    _TextConverter,
    _ValueConverter,
)
from cognite_toolkit._cdf_tk.utils.fileio import FileReader
from cognite_toolkit._cdf_tk.utils.fileio._readers import TableReader
from cognite_toolkit._cdf_tk.utils.http_client import DataBodyRequest, HTTPClient, HTTPMessage
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, TableUploadableStorageIO, UploadItem
from .selectors import DataPointsFileSelector


class DatapointsIO(TableUploadableStorageIO[DataPointsFileSelector, DataPointListResponse, DataPointInsertionRequest]):
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".csv"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    CHUNK_SIZE = 10_000
    BASE_SELECTOR = DataPointsFileSelector
    KIND = "Datapoints"
    SUPPORTED_READ_FORMATS = frozenset({".csv"})
    UPLOAD_ENDPOINT = "/timeseries/data"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = None

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._warned_columns: set[str] = set()
        self._epoc_converter = _EpochConverter(nullable=True)
        self._numeric_converter = _Float64Converter(nullable=True)
        self._string_converter = _TextConverter(nullable=True)

    def as_id(self, item: DataPointListResponse) -> str:
        raise NotImplementedError()

    def stream_data(self, selector: DataPointsFileSelector, limit: int | None = None) -> Iterable[Page]:
        raise NotImplementedError(f"Download of {type(DatapointsIO).__name__.removesuffix('IO')} is not yet supported")

    def count(self, selector: DataPointsFileSelector) -> int | None:
        raise NotImplementedError(f"Download of {type(DatapointsIO).__name__.removesuffix('IO')} is not yet supported")

    def data_to_json_chunk(
        self, data_chunk: Sequence[DataPointListResponse], selector: DataPointsFileSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise ToolkitNotImplementedError(
            f"Download of {type(DatapointsIO).__name__.removesuffix('IO')} does not support json format."
        )

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[DataPointInsertionRequest]],
        http_client: HTTPClient,
        selector: DataPointsFileSelector | None = None,
    ) -> Sequence[HTTPMessage]:
        results: list[HTTPMessage] = []
        for item in data_chunk:
            response = http_client.request_with_retries(
                DataBodyRequest(
                    endpoint_url=http_client.config.create_api_url(self.UPLOAD_ENDPOINT),
                    method="POST",
                    content_type="application/protobuf",
                    data_content=item.item.SerializeToString(),
                )
            )
            results.extend(response)
        return results

    def row_to_resource(
        self, source_id: str, row: dict[str, JsonVal], selector: DataPointsFileSelector | None = None
    ) -> DataPointInsertionRequest:
        if selector is None:
            raise ValueError("Selector must be provided to convert row to DataPointInsertionItem.")
        # We assume that the row was read using the read_chunks method.
        rows = cast(dict[str, list[Any]], row)
        if selector.timestamp_column not in rows:
            raise RuntimeError(f"Timestamp column '{selector.timestamp_column}' not found.")

        timestamps = list(
            self._convert_values(
                rows[selector.timestamp_column],
                self._epoc_converter,
                f"timestamps (columns {selector.timestamp_column!r})",
                source_id,
            )
        )

        datapoints_items: list[DataPointInsertionItem] = []
        for col, values in rows.items():
            if col == selector.timestamp_column:
                continue
            column = selector.id_by_column.get(col)
            if column is None:
                self._warn_missing_columns(col)
                continue
            args: dict[str, Any] = column.as_wrapped_id()
            if column.dtype == "numeric":
                number_values = self._convert_values(
                    values, self._numeric_converter, f"numeric datapoints (column {col!r})", source_id
                )
                args["numericDatapoints"] = NumericDatapoints(
                    datapoints=[
                        NumericDatapoint(timestamp=timestamp, value=value)
                        for timestamp, value in zip(timestamps, number_values)
                        if timestamp is not None
                    ]
                )
            elif column.dtype == "string":
                string_values = self._convert_values(
                    values, self._string_converter, f"string datapoints (column {col!r})", source_id
                )
                args["stringDatapoints"] = StringDatapoints(
                    datapoints=[
                        StringDatapoint(timestamp=timestamp, value=value)
                        for timestamp, value in zip(timestamps, string_values)
                        if timestamp is not None
                    ]
                )
            else:
                raise RuntimeError(f"Unsupported dtype {column.dtype} for column {col}.")

            datapoints_items.append(DataPointInsertionItem(**args))
        return DataPointInsertionRequest(items=datapoints_items)

    def _convert_values(
        self, values: list[Any], converter: _ValueConverter, name: str, source_id: str
    ) -> Iterable[Any]:
        failed_count = 0
        for value in values:
            try:
                converted = converter.convert(value)
            except ValueError:
                failed_count += 1
                continue
            yield converted
        if failed_count > 0:
            HighSeverityWarning(
                f"In {source_id}' {failed_count:,} {name} could not be converted and will be skipped."
            ).print_warning(console=self.client.console)

    def _warn_missing_columns(self, column: str) -> None:
        if column not in self._warned_columns:
            HighSeverityWarning(
                f"Column '{column}' not found in selector columns. Skipping this column."
            ).print_warning(console=self.client.console)
            self._warned_columns.add(column)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> DataPointInsertionRequest:
        raise ToolkitNotImplementedError(
            f"Upload of {type(DatapointsIO).__name__.removesuffix('IO')} does not support json format."
        )

    @classmethod
    def read_chunks(cls, reader: FileReader) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        if not isinstance(reader, TableReader):
            raise RuntimeError("DatapointsIO can only read from TableReader instances.")
        iterator = iter(reader.read_chunks_with_line_numbers())
        try:
            start_row, first = next(iterator)
        except StopIteration:
            # Empty file
            return
        column_names = list(first.keys())
        batch: dict[str, list[Any]] = {col: [value] for col, value in first.items()}
        last_row = start_row
        for row_no, chunk in iterator:
            for col, value in chunk.items():
                batch[col].append(value)

            # The number of datapoints is the number of rows times the number of value columns.
            if ((len(column_names) - 1) * len(batch[column_names[0]])) >= cls.CHUNK_SIZE:
                # We cannot guarantee JsonVal here, but that is handled later in the processing pipeline.
                yield [(f"rows {start_row} to {row_no}", batch)]  # type: ignore[list-item]
                start_row = row_no + 1
                batch = {col: [] for col in column_names}
            last_row = row_no
        if any(batch.values()):
            yield [(f"rows {start_row} to{last_row}", batch)]  # type: ignore[list-item]



import sys
from collections.abc import Callable, Collection, Iterator, MutableSequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, SupportsIndex, overload

from cognite_toolkit._cdf_tk.cruds import ResourceTypes
from cognite_toolkit._cdf_tk.cruds._base_cruds import T_ID

from ._build_variables import BuildVariables
from ._built_resources import (
    BuiltFullResourceList,
    BuiltResource,
    BuiltResourceList,
    SourceLocation,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class BuiltModule:
    name: str
    location: SourceLocation
    build_variables: BuildVariables
    resources: dict[str, BuiltResourceList]
    warning_count: int
    status: str
    iteration: int
    package_id: str | None = None
    module_id: str | None = None

    @classmethod
    def load(cls, data: dict[str, Any]) -> Self:
        return cls(
            name=data["name"],
            location=SourceLocation.load(data["location"]),
            build_variables=BuildVariables.load(data["build_variables"]),
            resources={
                key: BuiltResourceList([BuiltResource.load(resource_data, key) for resource_data in resources_data])
                for key, resources_data in data["resources"].items()
            },
            warning_count=data.get("warning_count", 0),
            status=data.get("status", "Success"),
            iteration=data.get("iteration", 1),
            package_id=data.get("package_id"),
            module_id=data.get("module_id"),
        )

    def dump(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "location": self.location.dump(),
            "build_variables": self.build_variables.dump(),
            "resources": {
                key: [resource.dump(key) for resource in resources] for key, resources in self.resources.items()
            },
            "warning_count": self.warning_count,
            "status": self.status,
            "iteration": self.iteration,
            "package_id": self.package_id,
            "module_id": self.module_id,
        }


@dataclass
class BuiltModuleList(list, MutableSequence[BuiltModule]):
    # Implemented to get correct type hints
    def __init__(self, collection: Collection[BuiltModule] | None = None) -> None:
        super().__init__(collection or [])

    def __iter__(self) -> Iterator[BuiltModule]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> BuiltModule: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "BuiltModule | BuiltModuleList":
        if isinstance(index, slice):
            return BuiltModuleList(super().__getitem__(index))
        return super().__getitem__(index)

    def get_resources(
        self,
        id_type: type[T_ID] | None,
        resource_dir: ResourceTypes,
        kind: str | None = None,
        selected: Path | str | None = None,
        is_supported_file: Callable[[Path], bool] | None = None,
    ) -> "BuiltFullResourceList[T_ID]":
        resources = (
            resource.create_full(module, resource_dir)
            for module in self
            for resource in module.resources.get(resource_dir, [])
            if kind is None or resource.kind == kind
        )
        if isinstance(selected, str):
            resources = (resource for resource in resources if resource.module_name == selected)
        elif isinstance(selected, Path):
            resources = (
                resource
                for resource in resources
                if (resource.source.path == selected or self._are_relative(resource.source.path, selected))
            )
        if is_supported_file:
            # This is necessary as the destination file can be created from a source file that is not supported.
            # This happens for RAW table files which produces a Database file.
            resources = (resource for resource in resources if is_supported_file(resource.source.path))

        return BuiltFullResourceList[T_ID](list(resources))

    def as_resources_by_folder(self) -> dict[str, BuiltResourceList[T_ID]]:
        resources_by_folder: dict[str, BuiltResourceList[T_ID]] = {}
        for module in self:
            for resource_dir, resources in module.resources.items():
                resources_by_folder.setdefault(resource_dir, BuiltResourceList()).extend(resources)
        return resources_by_folder

    @staticmethod
    def _are_relative(filepath: Path, select_path: Path) -> bool:
        return filepath.resolve().is_relative_to(select_path.resolve())



import gzip
from typing import Any

from cognite.client._api.transformations import TransformationsAPI
from cognite.client.config import global_config
from cognite.client.data_classes import TransformationPreviewResult
from cognite.client.utils import _json
from requests import Response


class FixedTransformationsAPI(TransformationsAPI):
    """This overrides the preview method in the transformations API to allow for longer timeouts and retries."""

    _default_timeout_run_query = 240  # seconds, this is the maximum timeout for running queries in CDF

    def preview(
        self,
        query: str | None = None,
        convert_to_string: bool = False,
        limit: int | None = 100,
        source_limit: int | None = 100,
        infer_schema_limit: int | None = 10_000,
        timeout: int | None = _default_timeout_run_query,
    ) -> TransformationPreviewResult:
        """`Preview the result of a query. <https://developer.cognite.com/api#tag/Query/operation/runPreview>`_

        This method overrides the default preview method in the TransformationsAPI to allow for longer
        timeouts and retries.

        Toolkit runs long-running queries that takes longer than the typical default of 30 seconds. In addition,
        we do not want to retry, which typically up to 10 times, as the user will have to wait for a long time. Instead,
        we want to fail provide the user with the error and then let the user decide whether to retry or not by
        running the CLI command again.

        Args:
            query (str | None): SQL query to run for preview.
            convert_to_string (bool): Stringify values in the query results, default is False.
            limit (int | None): Maximum number of rows to return in the final result, default is 100.
            source_limit (int | None): Maximum number of items to read from the data source or None to run without limit, default is 100.
            infer_schema_limit (int | None): Limit for how many rows that are used for inferring result schema, default is 10 000.
            timeout (int | None): Number of seconds to wait before cancelling a query. The default, and maximum, is 240.

        Returns:
            TransformationPreviewResult: Result of the executed query
        """
        request_body = {
            "query": query,
            "convertToString": convert_to_string,
            "limit": limit,
            "sourceLimit": source_limit,
            "inferSchemaLimit": infer_schema_limit,
            "timeout": timeout,
        }
        response = self._do_requests_retry(
            "POST",
            "/transformations/query/run",
            json=request_body,
            timeout=timeout or (self._default_timeout_run_query + 60),  # add a buffer to the timeout
            retry=False,
        )
        return TransformationPreviewResult._load(response.json(), cognite_client=self._cognite_client)

    def _do_requests_retry(
        self,
        method: str,
        url_path: str,
        accept: str = "application/json",
        api_subversion: str | None = None,
        retry: bool = True,
        **kwargs: Any,
    ) -> Response:
        """This is a copy of the _do_requests method from the APIClient with the option to turn off retries."""
        is_retryable, full_url = self._resolve_url(method, url_path)
        json_payload = kwargs.pop("json", None)
        headers = self._configure_headers(
            accept,
            additional_headers=self._config.headers.copy(),
            api_subversion=api_subversion,
        )
        headers.update(kwargs.get("headers") or {})

        if json_payload is not None:
            try:
                data = _json.dumps(json_payload, allow_nan=False)
            except ValueError as e:
                # A lot of work to give a more human friendly error message when nans and infs are present:
                msg = "Out of range float values are not JSON compliant"
                if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                    raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                        e.__traceback__
                    ) from None
                raise
            kwargs["data"] = data
            if method in ["PUT", "POST"] and not global_config.disable_gzip:
                kwargs["data"] = gzip.compress(data.encode())
                headers["Content-Encoding"] = "gzip"

        kwargs["headers"] = headers

        # requests will by default follow redirects. This can be an SSRF-hazard if
        # the client can be tricked to request something with an open redirect, in
        # addition to leaking the token, as requests will send the headers to the
        # redirected-to endpoint.
        # If redirects are to be followed in a call, this should be opted into instead.
        kwargs.setdefault("allow_redirects", False)

        if is_retryable and retry:
            res = self._http_client_with_retry.request(method=method, url=full_url, **kwargs)
        else:
            res = self._http_client.request(method=method, url=full_url, **kwargs)

        match res.status_code:
            case 200 | 201 | 202 | 204:
                pass
            case 401:
                self._raise_no_project_access_error(res)
            case _:
                self._raise_api_error(res, payload=json_payload)

        stream = kwargs.get("stream")
        self._log_request(res, payload=json_payload, stream=stream)
        return res



import sys
from collections.abc import ItemsView, Iterable, Iterator, KeysView, Mapping, MutableMapping, ValuesView
from dataclasses import dataclass, field
from pathlib import Path

from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.tk_warnings.base import ToolkitWarning, WarningList
from cognite_toolkit._cdf_tk.tk_warnings.other import LowSeverityWarning

from ._module_directories import ModuleDirectories, ModuleLocation

if sys.version_info >= (3, 11):
    from typing import Self

    import toml
else:
    import tomli as toml
    from typing_extensions import Self


@dataclass
class Package:
    """A package represents a bundle of modules.
    Args:
        name: the unique identifier of the package.
        title: The display name of the package.
        description: A description of the package.
        id: An optional identifier for package statistics.
        modules: The modules that are part of the package.
    """

    name: str
    title: str
    description: str | None = None
    id: str | None = None
    can_cherry_pick: bool = True
    modules: list[ModuleLocation] = field(default_factory=list)

    @property
    def module_names(self) -> set[str]:
        """The names of the modules in the package."""
        return {module.name for module in self.modules}

    @classmethod
    def load(cls, name: str, package_definition: dict) -> Self:
        return cls(
            name=name,
            title=package_definition["title"],
            description=package_definition.get("description"),
            id=package_definition.get("id"),
            can_cherry_pick=package_definition.get("canCherryPick", True),
        )


class Packages(dict, MutableMapping[str, Package]):
    warnings: WarningList[ToolkitWarning] | None = None

    def __init__(
        self,
        packages: Iterable[Package] | Mapping[str, Package] | None = None,
        warnings: WarningList[ToolkitWarning] | None = None,
    ) -> None:
        if packages is None:
            super().__init__()
        elif isinstance(packages, Mapping):
            super().__init__(packages)
        else:
            super().__init__({p.name: p for p in packages})

        if warnings:
            self.warnings = warnings

    @classmethod
    def load(
        cls,
        root_module_dir: Path,
    ) -> Self:
        """Loads the packages in the source directory.

        Args:
            root_module_dir: The module directories to load the packages from.
        """

        package_definition_path = next(root_module_dir.rglob("packages.toml"), None)
        if not package_definition_path or not package_definition_path.exists():
            raise ToolkitFileNotFoundError(f"Package manifest toml not found at {package_definition_path}")

        library_definition = toml.loads(package_definition_path.read_text(encoding="utf-8"))
        package_definitions = library_definition.get("packages", {})

        # Load all available modules
        module_directories = ModuleDirectories.load(root_module_dir)

        # Create lookup dictionaries for efficient module discovery
        module_by_relative_path = {module.relative_path: module for module in module_directories}

        packages_with_modules: dict[str, Package] = {}

        warnings = WarningList[ToolkitWarning]()

        for package_name, package_definition in package_definitions.items():
            packages_with_modules[package_name] = Package.load(package_name, package_definition)
            if modules := package_definition.get("modules"):
                if isinstance(modules, list) and modules:
                    for module_path in modules:
                        if (module := module_by_relative_path.get(Path(module_path))) is None:
                            warnings.append(
                                LowSeverityWarning(
                                    f"Unable to load module '{module_path}'. The path may be wrong or the module may require an alpha flag that is not set."
                                )
                            )
                            continue
                        packages_with_modules[package_name].modules.append(module)

        return cls(packages_with_modules, warnings)

    # The methods are overloads to provide type hints for the methods.
    def items(self) -> ItemsView[str, Package]:  # type: ignore[override]
        return super().items()

    def keys(self) -> KeysView[str]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[Package]:  # type: ignore[override]
        return super().values()

    def __iter__(self) -> Iterator[str]:
        yield from super().__iter__()

    def __getitem__(self, package_name: str) -> Package:
        return super().__getitem__(package_name)



import os
from datetime import datetime
from unittest.mock import patch

import pytest
from cognite.client.data_classes import ClientCredentials
from cognite.client.data_classes.functions import Function, FunctionCall
from cognite.client.data_classes.transformations import Transformation
from cognite.client.data_classes.workflows import (
    WorkflowExecution,
)

from cognite_toolkit._cdf_tk.commands import RunFunctionCommand, RunTransformationCommand, RunWorkflowCommand
from cognite_toolkit._cdf_tk.commands.run import FunctionCallArgs
from cognite_toolkit._cdf_tk.data_classes import ModuleResources
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import RUN_DATA
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestRunTransformation:
    def test_run_transformation(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        transformation = Transformation(
            name="Test transformation",
            external_id="test",
            query="SELECT * FROM timeseries",
        )
        toolkit_client_approval.append(Transformation, transformation)

        assert RunTransformationCommand().run_transformation(toolkit_client_approval.mock_client, "test") is True


@pytest.fixture(scope="session")
def functon_module_resources() -> ModuleResources:
    return ModuleResources(RUN_DATA, "dev")


class TestRunFunction:
    def test_run_function_live(
        self, toolkit_client_approval: ApprovalToolkitClient, env_vars_with_client: EnvironmentVariables
    ) -> None:
        function = Function(
            id=1234567890,
            name="test3",
            external_id="fn_test3",
            description="Returns the input data, secrets, and function info.",
            owner="pytest",
            status="RUNNING",
            file_id=1234567890,
            function_path="./handler.py",
            created_time=int(datetime.now().timestamp() / 1000),
            secrets={"my_secret": "***"},
        )
        toolkit_client_approval.append(Function, function)
        toolkit_client_approval.mock_client.functions.call.return_value = FunctionCall(
            id=1234567890,
            status="RUNNING",
            start_time=int(datetime.now().timestamp() / 1000),
        )
        cmd = RunFunctionCommand()

        cmd.run_cdf(
            env_vars_with_client,
            organization_dir=RUN_DATA,
            build_env_name="dev",
            external_id="fn_test3",
            data_source="daily-8pm-utc",
            wait=False,
        )
        assert toolkit_client_approval.mock_client.functions.call.called

    @patch.dict(
        os.environ,
        {
            "IDP_FUN_CLIENT_ID": "dummy",
            "IDP_FUN_CLIENT_SECRET": "dummy",
        },
    )
    def test_run_local_function(self, env_vars_with_client: EnvironmentVariables) -> None:
        cmd = RunFunctionCommand()

        cmd.run_local(
            env_vars=env_vars_with_client,
            organization_dir=RUN_DATA,
            build_env_name="dev",
            external_id="fn_test3",
            data_source="daily-8pm-utc",
            rebuild_env=False,
            virtual_env_folder_name="function_local_venvs_test_run_local_function",
        )

    @patch.dict(
        os.environ,
        {
            "IDP_WF_CLIENT_ID": "dummy",
            "IDP_WF_CLIENT_SECRET": "dummy",
        },
    )
    def test_run_local_function_with_workflow(self, env_vars_with_client: EnvironmentVariables) -> None:
        cmd = RunFunctionCommand()

        cmd.run_local(
            env_vars=env_vars_with_client,
            organization_dir=RUN_DATA,
            build_env_name="dev",
            external_id="fn_test3",
            data_source="workflow",
            rebuild_env=False,
            virtual_env_folder_name="function_local_venvs_test_run_local_function_workflow",
        )

    @pytest.mark.parametrize(
        "data_source, expected",
        [
            pytest.param(
                "workflow",
                FunctionCallArgs(
                    data={
                        "breakfast": "today: egg and bacon",
                        "lunch": "today: a chicken",
                        "dinner": "today: steak with stakes on the side",
                    },
                    authentication=ClientCredentials(
                        client_id="workflow_client_id",
                        client_secret="workflow_client_secret",
                    ),
                    client_id_env_name="IDP_WF_CLIENT_ID",
                    client_secret_env_name="IDP_WF_CLIENT_SECRET",
                ),
                id="workflow",
            ),
            pytest.param(
                "daily-8am-utc",
                FunctionCallArgs(
                    data={
                        "breakfast": "today: peanut butter sandwich and coffee",
                        "lunch": "today: greek salad and water",
                        "dinner": "today: steak and red wine",
                    },
                    authentication=ClientCredentials(
                        client_id="function_client_id",
                        client_secret="function_client_secret",
                    ),
                    client_id_env_name="IDP_FUN_CLIENT_ID",
                    client_secret_env_name="IDP_FUN_CLIENT_SECRET",
                ),
                id="daily-8pm-utc",
            ),
        ],
    )
    def test_get_call_args(
        self, data_source: str, expected: FunctionCallArgs, functon_module_resources: ModuleResources
    ) -> None:
        environment_variables = {
            expected.client_id_env_name: expected.authentication.client_id,
            expected.client_secret_env_name: expected.authentication.client_secret,
        }
        actual = RunFunctionCommand._get_call_args(
            data_source, "fn_test3", functon_module_resources, environment_variables, is_interactive=False
        )

        assert actual == expected


class TestRunWorkflow:
    def test_run_workflow(
        self, toolkit_client_approval: ApprovalToolkitClient, env_vars_with_client: EnvironmentVariables
    ):
        toolkit_client_approval.mock_client.workflows.executions.run.return_value = WorkflowExecution(
            id="1234567890",
            workflow_external_id="workflow",
            status="running",
            created_time=int(datetime.now().timestamp() / 1000),
            version="v1",
        )

        assert (
            RunWorkflowCommand().run_workflow(
                env_vars_with_client,
                organization_dir=RUN_DATA,
                build_env_name="dev",
                external_id="workflow",
                version="v1",
                wait=False,
            )
            is True
        )



from collections.abc import Callable, Iterable
from functools import partial
from pathlib import Path

from cognite.client.data_classes._base import T_CogniteResource

from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_STEM, DATA_RESOURCE_DIR
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import ConfigurableStorageIO, Page, StorageIO, T_Selector, TableStorageIO
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils.file import safe_write, sanitize_filename, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.fileio import TABLE_WRITE_CLS_BY_FORMAT, Compression, FileWriter, SchemaColumn
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import ToolkitCommand


class DownloadCommand(ToolkitCommand):
    def download(
        self,
        selectors: Iterable[T_Selector],
        io: StorageIO[T_Selector, T_CogniteResource],
        output_dir: Path,
        verbose: bool,
        file_format: str,
        compression: str,
        limit: int | None = 100_000,
    ) -> None:
        """Downloads data from CDF to the specified output directory.

        Args:
            selectors: The selectors of the resources to download.
            io: The StorageIO instance that defines how to download and process the data.
            output_dir: The directory where the downloaded files will be saved.
            verbose: If True, prints detailed information about the download process.
            file_format: The format of the files to be written (e.g., ".ndjson").
            compression: The compression method to use for the downloaded files (e.g., "none", "gzip").
            limit: The maximum number of items to download for each selected set. If None, all items will be downloaded.
        """
        compression_cls = Compression.from_name(compression)

        console = io.client.console
        for selector in selectors:
            target_dir = output_dir / sanitize_filename(selector.group)
            if verbose:
                console.print(f"Downloading {selector.display_name} '{selector!s}' to {target_dir.as_posix()!r}")

            iteration_count = self._get_iteration_count(io, selector, limit)
            filestem = sanitize_filename(str(selector))
            if self._already_downloaded(target_dir, filestem):
                warning = LowSeverityWarning(
                    f"Data for {selector!s} already exists in {target_dir.as_posix()!r}. Skipping download."
                )
                self.warn(warning, console=console)
                continue

            selector.dump_to_file(target_dir)
            columns: list[SchemaColumn] | None = None
            is_table = file_format in TABLE_WRITE_CLS_BY_FORMAT
            if is_table and isinstance(io, TableStorageIO):
                columns = io.get_schema(selector)
            elif is_table:
                raise ToolkitValueError(
                    f"Cannot download {selector.kind} in {file_format!r} format. The {selector.kind!r} storage type does not support table schemas."
                )

            with FileWriter.create_from_format(
                file_format, target_dir, selector.kind, compression_cls, columns=columns
            ) as writer:
                executor = ProducerWorkerExecutor[Page[T_CogniteResource], list[dict[str, JsonVal]]](
                    download_iterable=io.stream_data(selector, limit),
                    process=self.create_data_process(io=io, selector=selector, is_table=is_table),
                    write=partial(writer.write_chunks, filestem=filestem),
                    iteration_count=iteration_count,
                    # Limit queue size to avoid filling up memory before the workers can write to disk.
                    max_queue_size=8 * 10,  # 8 workers, 10 items per worker
                    download_description=f"Downloading {selector!s}",
                    process_description="Processing",
                    write_description=f"Writing to {target_dir.as_posix()!r} in files with stem {filestem!r}",
                    console=console,
                )
                executor.run()
                executor.raise_on_error()
                file_count = writer.file_count

            if isinstance(io, ConfigurableStorageIO):
                for config in io.configurations(selector):
                    filename = config.filename or filestem
                    config_file = target_dir / DATA_RESOURCE_DIR / config.folder_name / f"{filename}.{config.kind}.yaml"
                    config_file.parent.mkdir(parents=True, exist_ok=True)
                    safe_write(config_file, yaml_safe_dump(config.value))

            console.print(f"Downloaded {selector!s} to {file_count} file(s) in {target_dir.as_posix()!r}.")

    @staticmethod
    def _get_iteration_count(
        io: StorageIO[T_Selector, T_CogniteResource],
        selector: T_Selector,
        limit: int | None,
    ) -> int | None:
        total = io.count(selector)
        if total is not None and limit is not None and total > limit:
            total = limit
        iteration_count: int | None = None
        if total is not None:
            iteration_count = total // io.CHUNK_SIZE + (1 if total % io.CHUNK_SIZE > 0 else 0)
        return iteration_count

    @staticmethod
    def _already_downloaded(output_dir: Path, filestem: str) -> bool:
        if not output_dir.exists():
            return False

        # Check for multi-part files (e.g. ndjson, csv, parquet)
        if any(output_dir.glob(f"{filestem}-part-*")):
            return True

        # Check for single files (e.g. yaml) and exclude the metadata file.
        manifest_file_name = f"{filestem}.{DATA_MANIFEST_STEM}.yaml"
        for f in output_dir.glob(f"{filestem}.*"):
            if f.name != manifest_file_name:
                return True

        return False

    @staticmethod
    def create_data_process(
        io: StorageIO[T_Selector, T_CogniteResource],
        selector: T_Selector,
        is_table: bool,
    ) -> Callable[[Page[T_CogniteResource]], list[dict[str, JsonVal]]]:
        """Creates a data processing function based on the IO type and whether the output is a table."""
        if is_table and isinstance(io, TableStorageIO):

            def row_data_process(chunk: Page[T_CogniteResource]) -> list[dict[str, JsonVal]]:
                return io.data_to_row(chunk.items, selector)

            return row_data_process

        def chunk_data_process(data_page: Page[T_CogniteResource]) -> list[dict[str, JsonVal]]:
            return io.data_to_json_chunk(data_page.items, selector)

        return chunk_data_process



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.hosted_extractor_job import HostedExtractorJobYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_hosted_extractor_job_test_cases() -> Iterable:
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "invalid", "some_field": "some_value"},
            "config": {"topicFilter": "some_filter"},
        },
        {
            "In field format invalid type 'invalid'. Expected one of cognite, custom, rockwell or value",
        },
        id="Invalid type",
    )

    yield pytest.param(
        {
            "externalId": "myJob",
        },
        {
            "Missing required field: 'destinationId'",
            "Missing required field: 'sourceId'",
            "Missing required field: 'format'",
        },
        id="Missing required fields",
    )

    # Format validation - missing type
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"encoding": "utf8"},
        },
        {
            "In field format invalid input format missing 'type' key",
        },
        id="Format missing type",
    )

    # Format validation - invalid encoding for all JobFormat types
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite", "encoding": "invalid_encoding"},
        },
        {
            "In format.encoding input should be 'utf8', 'utf16', 'utf16le' or 'latin1'. Got 'invalid_encoding'.",
        },
        id="CogniteFormat invalid encoding",
    )

    # Format validation - invalid compression for all JobFormat types
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite", "compression": "invalid_compression"},
        },
        {
            "In format.compression input should be 'gzip'. Got 'invalid_compression'.",
        },
        id="CogniteFormat invalid compression",
    )

    # CustomFormat validation - missing mapping_id
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "custom"},
        },
        {
            "In format missing required field: 'mappingId'",
        },
        id="CustomFormat missing mappingId",
    )

    # CustomFormat validation - too long mapping_id
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "custom", "mappingId": "a" * 256},
        },
        {
            "In format.mappingId string should have at most 255 characters",
        },
        id="CustomFormat mappingId too long",
    )

    # DataModelFormat validation - prefix config for all DataModelFormat types
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite", "prefix": {"prefix": "a" * 256}},
        },
        {
            "In format.prefix.prefix string should have at most 255 characters",
        },
        id="CogniteFormat prefix too long",
    )

    # DataModelFormat validation - too many data models for all DataModelFormat types
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite", "dataModels": [{"space": f"space{i}"} for i in range(11)]},
        },
        {
            "In format.dataModels list should have at most 10 items after validation, not 11",
        },
        id="CogniteFormat too many data models",
    )

    # DataModelFormat validation - missing space in dataModels
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite", "dataModels": [{}]},
        },
        {
            "In format.dataModels[1] missing required field: 'space'",
        },
        id="CogniteFormat missing space in dataModels",
    )

    # Field length validation - externalId too long
    yield pytest.param(
        {
            "externalId": "a" * 256,
            "destinationId": "a" * 256,
            "sourceId": "a" * 256,
            "format": {"type": "cognite"},
        },
        {
            "In field externalId string should have at most 255 characters",
            "In field destinationId string should have at most 255 characters",
            "In field sourceId string should have at most 255 characters",
        },
        id="externalId too long",
    )

    # Format validation - non-dict format
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": "invalid_format",
        },
        {
            "In field format invalid input for format '<class 'str'>' expected dict",
        },
        id="Format non-dict type",
    )

    # Format validation - null format
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": None,
        },
        {
            "In field format invalid input for format '<class 'NoneType'>' expected dict",
        },
        id="Format null type",
    )

    # Rest Config validation - invalid IncrementalLoad type
    yield pytest.param(
        {
            "externalId": "myJob",
            "destinationId": "myDest",
            "sourceId": "mySource",
            "format": {"type": "cognite"},
            "config": {
                "interval": "1h",
                "path": "/some/path",
                "method": "get",
                "incrementalLoad": {"type": "nextUrl", "value": "some_value"},
                "pagination": {"type": "body", "value": "some_value"},
            },
        },
        {
            "In config.KafkaConfig missing required field: 'topic'",
            "In config.KafkaConfig unused field: 'incrementalLoad'",
            "In config.KafkaConfig unused field: 'interval'",
            "In config.KafkaConfig unused field: 'method'",
            "In config.KafkaConfig unused field: 'pagination'",
            "In config.KafkaConfig unused field: 'path'",
            "In config.MQTTConfig missing required field: 'topicFilter'",
            "In config.MQTTConfig unused field: 'incrementalLoad'",
            "In config.MQTTConfig unused field: 'interval'",
            "In config.MQTTConfig unused field: 'method'",
            "In config.MQTTConfig unused field: 'pagination'",
            "In config.MQTTConfig unused field: 'path'",
            "In config.RestConfig.incrementalLoad invalid type 'nextUrl'. Expected one of "
            "body, headerValue and queryParameter",
        },
        id="Invalid IncrementalLoad and Pagination type",
    )


class TestHostedExtractorJobYAML:
    @pytest.mark.parametrize("data", list(find_resources("Job", resource_dir="hosted_extractors")))
    def test_load_valid_hosted_extractor_job(self, data: dict[str, object]) -> None:
        loaded = HostedExtractorJobYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_hosted_extractor_job_test_cases()))
    def test_invalid_hosted_extractor_job_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, HostedExtractorJobYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import time
from collections.abc import Callable, Iterable
from typing import Any, Literal, NoReturn
from unittest.mock import MagicMock, patch

import pytest

from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor


@pytest.fixture
def setup_executor() -> tuple[
    ProducerWorkerExecutor, Callable[[list[int]], list[int]], Callable[[list[int]], None], int
]:
    download_iterable = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    process: MagicMock = MagicMock(side_effect=lambda x: [i * 2 for i in x])
    write_to_file: MagicMock = MagicMock()

    executor: ProducerWorkerExecutor = ProducerWorkerExecutor(
        download_iterable=download_iterable,
        process=process,
        write=write_to_file,
        iteration_count=len(download_iterable),
        max_queue_size=2,
    )
    return executor, process, write_to_file, len(download_iterable)


def test_run_success(setup_executor: tuple[ProducerWorkerExecutor, MagicMock, MagicMock, int]) -> None:
    executor, process, write_to_file, iteration_count = setup_executor
    executor.run()

    # Verify process was called for each batch
    assert process.call_count == iteration_count

    # Verify write_to_file was called for each processed batch
    assert write_to_file.call_count == iteration_count


def test_download_worker_handles_full_queue(monkeypatch: Any) -> None:
    max_queue_size = 1

    def process(items: list[int]) -> list[int]:
        # Simulate slow processing time
        time.sleep(1)
        return items

    def write_to_file(items: list[int]) -> None:
        nonlocal written
        written.append(items)

    to_download = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    written: list[list[int]] = []

    executor = ProducerWorkerExecutor(
        to_download, MagicMock(side_effect=process), MagicMock(side_effect=write_to_file), 2, max_queue_size
    )

    # Run the executor
    executor.run()

    assert written == to_download


class FailingIterator:
    def __init__(self, raise_in: Literal["iter", "next"] = "next") -> None:
        self.raise_in = raise_in

    def __iter__(self) -> "FailingIterator":
        if self.raise_in == "iter":
            raise TypeError("Error on iter call")
        return self

    def __next__(self) -> NoReturn:
        raise Exception("Error on first call")


def test_error_handling_in_download_worker() -> None:
    executor = ProducerWorkerExecutor(FailingIterator(raise_in="next"), MagicMock(), MagicMock(), 1, 10)
    executor.run()

    # Verify error occurred
    assert executor.error_occurred
    assert "Error on first call" in executor.error_message


def test_error_handling_in_download_worker_on_iter_call() -> None:
    executor = ProducerWorkerExecutor(FailingIterator(raise_in="iter"), MagicMock(), MagicMock(), 1, 10)
    executor.run()

    # Verify error occurred
    assert executor.error_occurred
    assert "Error on iter call" in executor.error_message


def test_error_handling_in_process_worker(
    setup_executor: tuple[ProducerWorkerExecutor, MagicMock, MagicMock, int],
) -> None:
    executor, process, _, _ = setup_executor
    process.side_effect = Exception("Process error")
    executor.run()

    # Verify error occurred
    assert executor.error_occurred
    assert "Process error" in executor.error_message


def test_error_handling_in_write_worker(
    setup_executor: tuple[ProducerWorkerExecutor, MagicMock, MagicMock, int],
) -> None:
    executor, _, write_to_file, _ = setup_executor
    write_to_file.side_effect = Exception("Write error")
    executor.run()

    # Verify error occurred
    assert executor.error_occurred
    assert "Write error" in executor.error_message


def test_kill_switch_stops_execution() -> None:
    downloaded: list[list[int]] = []
    to_download = [[1], [2], [3], [4], [5]]

    def slow_download() -> Iterable[list[int]]:
        for item in to_download:
            time.sleep(0.1)
            yield item
            downloaded.append(item)

    def user_input(timeout: float) -> str:
        return "q"

    with patch(f"{ProducerWorkerExecutor.__module__}.getch", user_input):
        executor = ProducerWorkerExecutor[list[list[int]], list[list[int]]](
            slow_download(), lambda x: x, lambda x: x, len(to_download), max_queue_size=2
        )
        executor.run()
        assert executor.stopped_by_user

    assert len(downloaded) < len(to_download)



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList, ChartWrite
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO
from .selectors import AllChartsSelector, CanvasSelector, ChartOwnerSelector, ChartSelector


class ChartIO(UploadableStorageIO[ChartSelector, Chart, ChartWrite]):
    KIND = "Charts"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = ChartSelector

    def as_id(self, item: Chart) -> str:
        return item.external_id

    def stream_data(self, selector: ChartSelector, limit: int | None = None) -> Iterable[Page]:
        selected_charts = self.client.charts.list(visibility="PUBLIC")
        if isinstance(selector, AllChartsSelector):
            ...
        elif isinstance(selector, ChartOwnerSelector):
            selected_charts = ChartList([chart for chart in selected_charts if chart.owner_id == selector.owner_id])
        else:
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for ChartIO")

        if limit is not None:
            selected_charts = ChartList(selected_charts[:limit])
        for chunk in chunker_sequence(selected_charts, self.CHUNK_SIZE):
            ts_ids_to_lookup = {
                ts_ref.ts_id
                for chart in chunk
                for ts_ref in chart.data.time_series_collection or []
                if ts_ref.ts_external_id is None and ts_ref.ts_id is not None
            }

            if ts_ids_to_lookup:
                retrieved_ts = self.client.time_series.retrieve_multiple(
                    ids=list(ts_ids_to_lookup), ignore_unknown_ids=True
                )
                id_to_external_id = {ts.id: ts.external_id for ts in retrieved_ts}

                for chart in chunk:
                    for ts_ref in chart.data.time_series_collection or []:
                        if ts_ref.ts_id in id_to_external_id:
                            ts_ref.ts_external_id = id_to_external_id[ts_ref.ts_id]
            yield Page(worker_id="main", items=chunk)

    def count(self, selector: ChartSelector) -> int | None:
        # There is no way to get the count of charts up front.
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[Chart], selector: ChartSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        return [chart.as_write().dump() for chart in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> ChartWrite:
        return ChartWrite._load(item_json)


class CanvasIO(UploadableStorageIO[CanvasSelector, IndustrialCanvas, IndustrialCanvasApply]):
    KIND = "IndustrialCanvas"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = CanvasSelector

    def as_id(self, item: IndustrialCanvas) -> str:
        return item.as_id()

    def stream_data(self, selector: CanvasSelector, limit: int | None = None) -> Iterable[Page]:
        raise ToolkitNotImplementedError("Streaming canvases is not implemented yet.")

    def count(self, selector: CanvasSelector) -> int | None:
        raise ToolkitNotImplementedError("Counting canvases is not implemented yet.")

    def data_to_json_chunk(
        self, data_chunk: Sequence[IndustrialCanvas], selector: CanvasSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        # Need to do lookup to get external IDs for all asset-centric resources.
        raise ToolkitNotImplementedError("Exporting canvases is not implemented yet.")

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> IndustrialCanvasApply:
        # Need to do lookup to get external IDs for all asset-centric resources.
        raise ToolkitNotImplementedError("Importing canvases is not implemented yet.")



from typing import Generic, TypeVar

from pydantic import BaseModel, Field, JsonValue

T = TypeVar("T", bound=BaseModel)


class PagedResponse(BaseModel, Generic[T]):
    items: list[T]
    next_cursor: str | None = Field(None, alias="nextCursor")


class QueryResponse(BaseModel, Generic[T]):
    items: dict[str, list[T]]
    typing: dict[str, JsonValue] | None = None
    next_cursor: dict[str, str] = Field(alias="nextCursor")
    debug: dict[str, JsonValue] | None = None



from cognite.client.data_classes import Annotation, AnnotationList, FileMetadata

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.storageio import FileAnnotationIO
from cognite_toolkit._cdf_tk.storageio.selectors import DataSetSelector
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestAnnotationIO:
    def test_stream_file_annotations(self) -> None:
        with monkeypatch_toolkit_client() as toolkit_client:
            # Allow lookup of internal IDs to external IDs
            approval_client = ApprovalToolkitClient(toolkit_client, allow_reverse_lookup=True)
            approval_client.mock_client.files.return_value = [[FileMetadata(external_id="file1", id=1)]]
            annotations = AnnotationList(
                [
                    Annotation(
                        annotation_type="diagrams.FileLink",
                        status="approved",
                        creating_app="test_app",
                        creating_app_version="v1",
                        annotated_resource_id=1,
                        annotated_resource_type="file",
                        creating_user="test_user",
                        data={
                            "fileRef": {
                                "id": 2,
                            },
                            "textRegion": {
                                "xMin": 0,
                                "xMax": 100,
                                "yMin": 0,
                                "yMax": 100,
                            },
                        },
                    ),
                    Annotation(
                        annotation_type="diagrams.AssetLink",
                        status="approved",
                        creating_app="test_app",
                        creating_app_version="v1",
                        annotated_resource_id=1,
                        annotated_resource_type="file",
                        creating_user="test_user",
                        data={
                            "assetRef": {"id": 3},
                            "textRegion": {
                                "xMin": 10,
                                "xMax": 110,
                                "yMin": 10,
                                "yMax": 110,
                            },
                        },
                    ),
                ]
            )
            approval_client.mock_client.annotations.list.return_value = annotations

            annotation_io = FileAnnotationIO(approval_client.client)
            selector = DataSetSelector(kind="FileMetadata", data_set_external_id="test_data_set")

            pages = list(annotation_io.stream_data(selector))
            assert len(pages) == 1
            total_items = sum(len(page.items) for page in pages)
            assert total_items == 2
            assert annotation_io.count(selector) is None

            json_data = annotation_io.data_to_json_chunk(pages[0].items)
            assert len(json_data) == 2
            assert json_data == [
                {
                    "annotatedResourceExternalId": "external_1",
                    "annotatedResourceType": "file",
                    "annotationType": "diagrams.FileLink",
                    "creatingApp": "test_app",
                    "creatingAppVersion": "v1",
                    "creatingUser": "test_user",
                    "data": {
                        "fileRef": {"externalId": "external_2"},
                        "textRegion": {"xMax": 100, "xMin": 0, "yMax": 100, "yMin": 0},
                    },
                    "status": "approved",
                },
                {
                    "annotatedResourceExternalId": "external_1",
                    "annotatedResourceType": "file",
                    "annotationType": "diagrams.AssetLink",
                    "creatingApp": "test_app",
                    "creatingAppVersion": "v1",
                    "creatingUser": "test_user",
                    "data": {
                        "assetRef": {"externalId": "external_3"},
                        "textRegion": {"xMax": 110, "xMin": 10, "yMax": 110, "yMin": 10},
                    },
                    "status": "approved",
                },
            ]



from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Any, Literal

from cognite.client.data_classes.data_modeling import EdgeId, NodeId, ViewId
from cognite.client.utils._identifier import InstanceId
from pydantic import Field

from cognite_toolkit._cdf_tk.constants import DM_EXTERNAL_ID_PATTERN, DM_VERSION_PATTERN, SPACE_FORMAT_PATTERN
from cognite_toolkit._cdf_tk.storageio._data_classes import InstanceIdCSVList

from ._base import DataSelector, SelectorObject


class SelectedView(SelectorObject):
    space: str = Field(
        description="Id of the space that the view belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the view.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str | None = Field(
        None,
        description="Version of the view.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )

    def as_id(self) -> ViewId:
        return ViewId(space=self.space, external_id=self.external_id, version=self.version)

    def __str__(self) -> str:
        if self.version:
            return f"{self.space}_{self.external_id}_{self.version}"
        return f"{self.space}_{self.external_id}"


class InstanceSelector(DataSelector, ABC):
    kind: Literal["Instances"] = "Instances"

    @abstractmethod
    def get_schema_spaces(self) -> list[str] | None:
        raise NotImplementedError()

    @abstractmethod
    def get_instance_spaces(self) -> list[str] | None:
        raise NotImplementedError()


class InstanceSpaceSelector(InstanceSelector):
    type: Literal["instanceSpace"] = "instanceSpace"
    instance_space: str
    instance_type: Literal["node", "edge"] = "node"
    view: SelectedView | None = None

    def get_schema_spaces(self) -> list[str] | None:
        return [self.view.space] if self.view else None

    def get_instance_spaces(self) -> list[str] | None:
        return [self.instance_space]

    @property
    def group(self) -> str:
        return self.instance_space

    def __str__(self) -> str:
        if self.view is None:
            return self.instance_type
        return f"{self.view}_{self.instance_type}"

    def as_filter_args(self) -> dict[str, Any]:
        args: dict[str, Any] = {
            "instance_type": self.instance_type,
            "space": self.instance_space,
        }
        if self.view:
            args["source"] = self.view.as_id()
        return args


class InstanceViewSelector(InstanceSelector):
    type: Literal["instanceView"] = "instanceView"
    view: SelectedView
    instance_type: Literal["node", "edge"] = "node"
    instance_spaces: tuple[str, ...] | None = None

    def get_schema_spaces(self) -> list[str] | None:
        return [self.view.space]

    def get_instance_spaces(self) -> list[str] | None:
        return list(self.instance_spaces) if self.instance_spaces else None

    @property
    def group(self) -> str:
        return self.view.space

    def __str__(self) -> str:
        return f"{self.view.external_id}_{self.view.version}_{self.instance_type}"

    def as_filter_args(self) -> dict[str, Any]:
        args: dict[str, Any] = {
            "instance_type": self.instance_type,
            "source": self.view.as_id(),
        }
        if self.instance_spaces:
            args["space"] = list(self.instance_spaces)
        return args


class InstanceFileSelector(InstanceSelector):
    type: Literal["instanceFile"] = "instanceFile"

    datafile: Path
    validate_instance: bool = True

    @property
    def group(self) -> str:
        return "Instances"

    def __str__(self) -> str:
        return f"file_{self.datafile.as_posix()}"

    @cached_property
    def items(self) -> InstanceIdCSVList:
        return InstanceIdCSVList.read_csv_file(self.datafile)

    @cached_property
    def _ids_by_type(self) -> tuple[list[NodeId], list[EdgeId]]:
        node_ids: list[NodeId] = []
        edge_ids: list[EdgeId] = []
        for instance in self.items:
            if instance.instance_type == "node":
                node_ids.append(NodeId(instance.space, instance.external_id))
            else:
                edge_ids.append(EdgeId(instance.space, instance.external_id))
        return node_ids, edge_ids

    @property
    def instance_ids(self) -> list[InstanceId]:
        node_ids, edge_ids = self._ids_by_type
        return [*node_ids, *edge_ids]

    @property
    def node_ids(self) -> list[NodeId]:
        return self._ids_by_type[0]

    @property
    def edge_ids(self) -> list[EdgeId]:
        return self._ids_by_type[1]

    def get_schema_spaces(self) -> list[str] | None:
        return None

    def get_instance_spaces(self) -> list[str] | None:
        return sorted({instance.space for instance in self.items})



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes import (
    Relationship,
    RelationshipList,
    RelationshipWrite,
    RelationshipWriteList,
    capabilities,
)
from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import RelationshipYAML

from .classic import AssetCRUD, EventCRUD, SequenceCRUD
from .data_organization import DataSetsCRUD, LabelCRUD
from .file import FileMetadataCRUD
from .timeseries import TimeSeriesCRUD


@final
class RelationshipCRUD(ResourceCRUD[str, RelationshipWrite, Relationship, RelationshipWriteList, RelationshipList]):
    folder_name = "classic"
    filename_pattern = r"^.*\.Relationship$"  # Matches all yaml files whose stem ends with '.Asset'.
    filetypes = frozenset({"yaml", "yml"})
    resource_cls = Relationship
    resource_write_cls = RelationshipWrite
    list_cls = RelationshipList
    list_write_cls = RelationshipWriteList
    kind = "Relationship"
    yaml_cls = RelationshipYAML
    dependencies = frozenset(
        {DataSetsCRUD, AssetCRUD, EventCRUD, SequenceCRUD, FileMetadataCRUD, TimeSeriesCRUD, LabelCRUD}
    )
    _doc_url = "Relationships/operation/createRelationships"

    @property
    def display_name(self) -> str:
        return "relationships"

    @classmethod
    def get_id(cls, item: Relationship | RelationshipWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Relationship must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RelationshipWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.RelationshipsAcl.Scope.All | capabilities.RelationshipsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.RelationshipsAcl.Scope.All()
        )

        actions = (
            [capabilities.RelationshipsAcl.Action.Read]
            if read_only
            else [capabilities.RelationshipsAcl.Action.Read, capabilities.RelationshipsAcl.Action.Write]
        )

        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.RelationshipsAcl.Scope.DataSet(list(data_set_ids))

        return capabilities.RelationshipsAcl(actions, scope)

    def create(self, items: RelationshipWriteList) -> RelationshipList:
        return self.client.relationships.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> RelationshipList:
        return self.client.relationships.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: RelationshipWriteList) -> RelationshipList:
        return self.client.relationships.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.relationships.delete(external_id=ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                self.client.relationships.delete(external_id=existing)
            return len(existing)
        else:
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Relationship]:
        return iter(
            self.client.relationships(data_set_external_ids=[data_set_external_id] if data_set_external_id else None)
        )

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        for label in item.get("labels", []):
            if isinstance(label, dict):
                yield LabelCRUD, label["externalId"]
            elif isinstance(label, str):
                yield LabelCRUD, label
        for connection in ["source", "target"]:
            type_key = f"{connection}Type"
            id_key = f"{connection}ExternalId"
            if type_key in item and id_key in item:
                type_value = item[type_key]
                id_value = item[id_key]
                if isinstance(id_value, str) and isinstance(type_value, str):
                    type_value = type_value.strip().casefold()
                    if type_value == "asset":
                        yield AssetCRUD, id_value
                    elif type_value == "sequence":
                        yield SequenceCRUD, id_value
                    elif type_value == "timeseries":
                        yield TimeSeriesCRUD, id_value
                    elif type_value == "file":
                        yield FileMetadataCRUD, id_value
                    elif type_value == "event":
                        yield EventCRUD, id_value

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> RelationshipWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return RelationshipWrite._load(resource)

    def dump_resource(self, resource: Relationship, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if not dumped.get("labels") and "labels" not in local:
            dumped.pop("labels", None)
        return dumped



from typing import Any, Literal

from cognite.client.data_classes import TransformationWrite
from pydantic import Field, field_validator, model_serializer
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from .authentication import AuthenticationClientIdSecret, OIDCCredential
from .base import ToolkitResource
from .transformation_destination import Destination


class TransformationYAML(ToolkitResource):
    _cdf_resource = TransformationWrite
    external_id: str = Field(description="The external ID provided by the client.")
    name: str = Field(description="Name of the transformation.")
    ignore_null_fields: bool = Field(
        description="Indicates how null values are handled on updates: ignore or set null."
    )
    destination: Destination | None = Field(default=None, description="Destination data type.")
    query: str | None = Field(default=None, description="SQL query of the transformation.")
    conflict_mode: Literal["abort", "delete", "update", "upsert"] | None = Field(
        default=None,
        description="Behavior when the data already exists.",
    )
    is_public: bool | None = Field(
        default=None,
        description="Indicates if the transformation is visible to all in project or only to the owner.",
    )
    authentication: (
        AuthenticationClientIdSecret
        | OIDCCredential
        | dict[Literal["read", "write"], AuthenticationClientIdSecret | OIDCCredential]
        | None
    ) = Field(
        default=None,
        description="Authentication information for the transformation.",
    )
    data_set_external_id: str | None = Field(
        default=None,
        description="External ID of the data set to which the transformation belongs.",
    )
    tags: list[str] | None = Field(
        default=None,
        description="List of tags for the Transformation.",
        max_length=5,
    )
    queryFile: str | None = Field(
        default=None,
        description="Used by Toolkit: Path to the SQL file containing the query for the transformation.",
    )

    @model_serializer(mode="wrap")
    def serialize_transformation(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        serialized_data = handler(self)
        if self.destination:
            serialized_data["destination"] = self.destination.model_dump(**vars(info))
        return serialized_data

    @field_validator("authentication", mode="before")
    @classmethod
    def validate_serialization(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "read" in value or "write" in value:
            return {
                k: AuthenticationClientIdSecret.model_validate(v) if isinstance(v, dict) else v
                for k, v in value.items()
            }
        if "scopes" in value or "tokenUri" in value or "cdfProjectName" in value or "audience" in value:
            return OIDCCredential.model_validate(value)
        return AuthenticationClientIdSecret.model_validate(value)



import sys
from dataclasses import dataclass
from typing import Any, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class ProjectStatus(CogniteResource):
    """Project status information."""

    data_modeling_status: Literal["HYBRID", "DATA_MODELING_ONLY"]
    url_name: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            data_modeling_status=resource["dataModelingStatus"],
            url_name=resource["urlName"],
        )


class ProjectStatusList(CogniteResourceList[ProjectStatus]):
    """List of ProjectStatus items."""

    _RESOURCE = ProjectStatus

    @property
    def this_project(self) -> ProjectStatus:
        """Returns the ProjectStatus of the current project."""
        project = self._cognite_client.config.project
        for item in self:
            if item.url_name == project:
                return item
        raise ValueError(f"Project '{project}' not found in the list of projects.")



from typing import Any, Literal

from pydantic import Field

from .base import ToolkitResource


class NodeId(ToolkitResource):
    external_id: str
    space: str


class FileMetadataYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    name: str = Field(
        description="The name of the file.",
        max_length=256,
    )
    directory: str | None = Field(
        default=None,
        description="Directory associated with the file.",
        max_length=512,
    )
    instance_id: NodeId | None = Field(
        default=None,
        description="The Instance ID for the file when created in DMS.",
    )
    source: str | None = Field(
        default=None,
        description="The source of the file.",
        max_length=128,
    )
    mime_type: str | None = Field(
        default=None,
        description="File type. E.g., text/plain, application/pdf, etc.",
        max_length=256,
    )
    metadata: dict[str, str] | None = Field(
        default=None, description="Custom, application-specific metadata.", max_length=256
    )
    asset_external_ids: list[str] | None = Field(default=None, description="Asset IDs.", min_length=1, max_length=1000)
    data_set_external_id: str | None = Field(default=None, description="The dataSet Id for the item.")
    labels: list[dict[Literal["externalId"], str]] | None = Field(
        default=None,
        description="A list of labels associated with this resource.",
        max_length=10,
    )
    geo_location: dict[str, Any] | None = Field(
        default=None,
        description="The geographic metadata of the file.",
    )
    source_created_time: int | None = Field(
        default=None, description="The timestamp for when the file was originally created in the source system.", ge=0
    )
    source_modified_time: int | None = Field(
        default=None, description="The timestamp for when the file was last modified in the source system.", ge=0
    )
    security_categories: list[str] | None = Field(
        default=None, description="The security category IDs required to access this file.", max_length=100
    )



from pathlib import Path
from typing import Annotated, Any

import typer

from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.constants import DATA_DEFAULT_DIR
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

DEFAULT_INPUT_DIR = Path.cwd() / DATA_DEFAULT_DIR


class UploadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.upload_main)

    @staticmethod
    def upload_main(
        ctx: typer.Context,
        input_dir: Annotated[
            Path,
            typer.Argument(
                help="The directory containing the data to upload.",
                exists=True,
                file_okay=False,
                dir_okay=True,
                resolve_path=True,
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the command will not actually upload the data, but will print what would be uploaded.",
            ),
        ] = False,
        deploy_resources: Annotated[
            bool,
            typer.Option(
                "--deploy-resources",
                "-r",
                help="If set, the command will look for resource configuration files in adjacent folders and create them if they do not exist.",
            ),
        ] = True,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Commands to upload data to CDF."""
        cmd = UploadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.upload(
                input_dir=input_dir,
                dry_run=dry_run,
                verbose=verbose,
                deploy_resources=deploy_resources,
                client=client,
            )
        )
